import copy
import re

import numpy as np

from ..io.file import FileIOMixin

__all__ = ["ScalarFormatter", "ParameterFormatter", "FunctionFormatter",
           "ModelFunctionFormatter", "CostFunctionFormatter", "latexify_ascii"]

LATEX_GREEK_LETTERS = [
    "alpha", "beta", "gamma", "Gamma", "delta", "Delta", "epsilon", "zeta", "eta", "theta", "Theta",
    "vartheta", "iota", "kappa", "varlambda", "Lambda", "mu", "nu", "omicron", "pi", "Pi", "rho",
    "sigma", "Sigma", "tau", "upsilon", "Upsilon", "phi", "Phi", "varphi", "chi", "psi", "Psi",
    "omega", "Omega"
]

# -- formatters for model parameters and model functions
def latexify_ascii(ascii_string):
    """Create a true type latex string of an standard ascii string.

    :param ascii_string: The string to be converted
    :type ascii_string: str
    :rtype: str
    """
    if ascii_string is None:
        return None
    if len(ascii_string) == 1:
        return r"{%s}" % ascii_string
    if ascii_string.startswith("{") and ascii_string.endswith("}"):
        return ascii_string
    elif re.compile("^[\d\w]_[\d\w]$").match(ascii_string):
        return r"{%s}" % ascii_string
    _greek_letter_string = None
    for _greek_letter in LATEX_GREEK_LETTERS:
        if ascii_string == _greek_letter:
            _greek_letter_string = r"{\%s}" % _greek_letter
            break
        if re.compile("^%s_[\d\w]$" % _greek_letter).match(ascii_string):
            _greek_letter_string = r"{\%s}" % ascii_string
            break
        if re.compile("^[\d\w]_%s$" % _greek_letter).match(ascii_string):
            _greek_letter_string = r"{%s_\%s}" % (ascii_string[:1], _greek_letter)
            break
        if ascii_string.startswith(_greek_letter):
            for _greek_letter_2 in LATEX_GREEK_LETTERS:
                if ascii_string == _greek_letter + "_" + _greek_letter_2:
                    _greek_letter_string = r"{\%s_\%s}" % (_greek_letter, _greek_letter_2)
                    break
            if _greek_letter_string is not None:
                break

    if _greek_letter_string is not None:
        return re.sub("varlambda", "lambda", _greek_letter_string)

    _lpn = ascii_string.replace('_', r"\_")
    return r"{\tt %s}" % _lpn


# Naming convention: Arguments describe all arguments of a function, parameters only the fitted
# parameters excluding the independent variable(s)


class ScalarFormatter(object):
    def __init__(self, sigma, n_significant_digits=2):
        """Format a scalar to a specified precision, according to the uncertainty.

        :param float sigma: The uncertainty of the parameter.
        :param int n_significant_digits: Number of significant digits.
        """
        self._sigma = sigma
        self._n_significant_digits = n_significant_digits
        _sig = int(-np.floor(np.log10(self._sigma))) + self._n_significant_digits - 1
        # inner rounding needed for errors like 0.9999999 -> 1.0 (shift in decimal place)
        self._sig = int(-np.floor(np.log10(np.around(self._sigma, _sig)))) + self._n_significant_digits - 1

    def __call__(self, x):
        """Format the input to the precision given by the uncertainty.

        :param float x: The value to format.
        :rtype: str
        """
        # needed e.g. when rounding values like 9.999999 -> 10.0 (shift in decimal place)
        _rounded_x = abs(np.around(x, self._sig))
        # fallback to rounding to 10^(-1) if value is zero
        _log_abs_x = -1
        if _rounded_x:
            _log_abs_x = np.log10(np.abs(_rounded_x))
        _val_sig = int(self._sig - int(-np.floor(_log_abs_x)) + self._n_significant_digits - 1)

        _val_sig = max(_val_sig, 0)

        _template = "%#.{significance}g".format(significance=_val_sig)
        return _template % x


class ParameterFormatter(FileIOMixin, object):
    """Formatter class for model parameter objects.

    These objects store the relevant information for constructing plain-text and/or LaTeX string
    representations of model function parameters.

    For this, the original argument name, the name for representation, formatted as a
    plain-text/LaTeX string, its value and its
    uncertainty is stored.

    The formatted string is obtained by calling the :py:meth:`~.get_formatted` method.
    """

    def __init__(self, arg_name, value=None, error=None, asymmetric_error=None, name=None,
                 latex_name=None):
        """Construct a Parameter Formatter.

        :param str arg_name: A plain string indicating the parameter's signature inside the
            function call.
        :param value: The parameter value.
        :type value: float or None
        :param error: The symmetric parameter error.
        :type error: float or None
        :param asymmetric_error: The asymmetric parameter errors.
        :type asymmetric_error: tuple[float, float] or None
        :param name: A plain-text-formatted string indicating the parameter name.
        :type name: str or None
        :param latex_name: A LaTeX-formatted string indicating the parameter name.
        :type latex_name: str or None
        :rtype: ParameterFormatter
        """
        super(ParameterFormatter, self).__init__()
        self._arg_name = arg_name
        self.value = value
        self.error = error
        self.asymmetric_error = asymmetric_error

        self.name = name
        self.latex_name = latex_name  # latex_name setter requires self._name to be set beforehand

        self._fixed = False

    @classmethod
    def _get_base_class(cls):
        return ParameterFormatter

    @classmethod
    def _get_object_type_name(cls):
        return 'parameter_formatter'

    @property
    def arg_name(self):
        """Name of the function argument this formatter represents.

        :rtype: str
        """
        return self._arg_name

    @property
    def name(self):
        """The plain-text-formatted string indicating the parameter name.

        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            self._name = self.arg_name
        else:
            self._name = name

    @property
    def latex_name(self):
        """The LaTeX-formatted string indicating the parameter name.

        :rtype: str
        """
        return self._latex_name

    @latex_name.setter
    def latex_name(self, new_latex_name):
        # TODO: validate
        if new_latex_name is None:
            self._latex_name = latexify_ascii(self.name)
        elif new_latex_name.startswith('{') and new_latex_name.endswith('}'):
            self._latex_name = new_latex_name
        else:
            self._latex_name = '{' + new_latex_name + '}'

    @property
    def value(self):
        """The parameter value.

        :rtype: float or None"""
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def error(self):
        """The symmetric parameter error.

        :rtype: float or None"""
        return self._error

    @error.setter
    def error(self, err_spec):
        self._error = err_spec

    @property
    def error_rel(self):
        """The relative parameter error.

        :rtype: float or None"""
        if self._error is None:
            return None
        return self._error / self._value

    @property
    def asymmetric_error(self):
        """Tuple containing the asymmetric parameter errors.

        :rtype: tuple[float, float] or None"""
        return self._asymmetric_error

    @asymmetric_error.setter
    def asymmetric_error(self, err_spec):
        self._asymmetric_error = err_spec

    @property
    def error_up(self):
        """The upper uncertainty (only for asymmetric errors).

        :rtype: float or None
        """
        if self.asymmetric_error is None:
            return None
        return self.asymmetric_error[1]

    @property
    def error_down(self):
        """The lower uncertainty (only for asymmetric errors).

        :rtype: float or None
        """
        if self.asymmetric_error is None:
            return None
        return self.asymmetric_error[0]

    @property
    def fixed(self):
        """If the parameter has been fixed by the user. :py:obj:`True` when it's fixed,
        :py:obj:`False` when not.

        :rtype: bool
        """
        return self._fixed

    @fixed.setter
    def fixed(self, fixed):
        self._fixed = fixed

    def get_formatted(self, with_name=False, with_value=True, with_errors=True,
                      n_significant_digits=2, round_value_to_error=True, asymmetric_error=False,
                      format_as_latex=False):
        """Get a formatted string representing this model parameter.

        :param bool with_name: If :py:obj:`True`, the output will include the parameter name.
        :param bool with_value: If :py:obj:`True`, the output will include the parameter value.
        :param bool with_errors: If :py:obj:`True`, the output will include the parameter error(s).
        :param int n_significant_digits: Number of significant digits for rounding.
        :param bool round_value_to_error: If :py:obj:`True`, the parameter value will be rounded to
            the same precision as the uncertainty.
        :param bool asymmetric_error: If :py:obj:`True`, the asymmetric parameter uncertainties are
            used.
        :param bool format_as_latex: If :py:obj:`True`, the returned string will be formatted using
            LaTeX syntax.
        :return: The string representation of the parameter.
        :rtype: str
        """
        _display_string = ""
        if with_name:
            if format_as_latex:
                _display_string += "$%s$" % (self.latex_name,)
            else:
                _display_string += "%s" % (self.name,)

        if with_value:
            if with_name:
                _display_string += " = "

            if self.fixed:
                if format_as_latex:
                    _display_string += r"$%g$ (fixed)" % self._value
                else:
                    _display_string += "%g (fixed)" % self._value
            elif not with_errors or (not asymmetric_error and self.error in (None, 0)) or \
                    (asymmetric_error and (
                            self.asymmetric_error is None or np.all(self.asymmetric_error == 0))):
                if format_as_latex:
                    _display_string += f"$%.{n_significant_digits}g$" % self.value
                else:
                    _display_string += f"%.{n_significant_digits}g" % self.value
            else:
                if asymmetric_error:
                    _min_err = min(abs(self.error_up), abs(self.error_down))
                else:
                    _min_err = self.error
                # fallback to rounding to 10^(-1) if error is zero
                if not _min_err or np.isnan(_min_err):
                    _min_err = 1e-1

                # calculate decimal precision if rounding:
                if round_value_to_error:
                    val_formatter = ScalarFormatter(_min_err, n_significant_digits=n_significant_digits)
                    _val = val_formatter(self.value)
                    _err = "%#.{n}g".format(n=n_significant_digits) % self.error
                    if asymmetric_error:  # needed for different powers of 10 in asymmetric errs
                        if abs(self.error_down) <= abs(self.error_up):
                            _err_u = val_formatter(abs(self.error_up))
                            _err_d = "%#.{n}g".format(n=n_significant_digits) % abs(self.error_down)
                        else:
                            _err_d = val_formatter(abs(self.error_down))
                            _err_u = "%#.{n}g".format(n=n_significant_digits) % abs(self.error_up)
                # default cases if no rounding
                else:
                    _val = f"%.{n_significant_digits}g" % self.value
                    _err = f"%.{n_significant_digits}g" % self.error
                    if asymmetric_error:
                        _err_u = f"%.{n_significant_digits}g" % self.error_up
                        _err_d = f"%.{n_significant_digits}g" % self.error_down

                if asymmetric_error:
                    if format_as_latex:
                        _display_string += "${%s}^{+%s}_{-%s}$" % (_val, _err_u, _err_d)
                    else:
                        _display_string += "%s + %s (up) - %s (down)" % (_val, _err_u, _err_d)
                else:
                    if format_as_latex:
                        _display_string += "$%s \\pm %s$" % (_val, _err)
                    else:
                        _display_string += "%s +/- %s" % (_val, _err)

            # replace scientific notation with power of ten (LaTeX only)
            if format_as_latex:
                _display_string = re.sub(r'(-?\d*\.?\d+?)0*e\+?(-?[0-9]*[1-9]?)',
                                         r'\1\\times10^{\2}', _display_string)
        return _display_string


class FunctionFormatter(FileIOMixin, object):
    """Base class for function formatter objects. Requires further specialization for each type of
    model function.
    Objects derived from this class store information relevant for constructing plain-text and/or
    LaTeX string representations of functions.

    For this, the function name, formatted as a plain-text/LaTeX string, as well as a list of
    references to :py:obj:`ParameterFormatter` objects which contain information on how to format
    the model function arguments is stored.

    Optionally, plain-text/LaTeX expression strings can be provided. These are strings representing
    the model function expression (i.e. mathematical formula).

    The formatted string is obtained by calling the :py:meth:`~get_formatted` method.
    """
    DEFAULT_EXPRESSION_STRING = None
    DEFAULT_LATEX_EXPRESSION_STRING = None

    def __init__(self, name, latex_name=None, arg_formatters=None, expression_string=None,
                 latex_expression_string=None):
        """Construct a formatter for a model function:

        :param name: A plain-text-formatted string indicating the function name.
        :type name: str
        :param latex_name: A LaTeX-formatted string indicating the function name.
        :type latex_name: str
        :param arg_formatters: List of :py:obj:`ParameterFormatter`-derived objects, formatters for
            function arguments.
        :type arg_formatters: list[kafe2.fit._base.ParameterFormatter]
        :param expression_string: A plain-text-formatted string indicating the function expression.
        :type expression_string: str
        :param latex_expression_string: A LaTeX-formatted string indicating the function expression.
        :type latex_expression_string: str
        """
        # TODO should name be allowed to be None?
        self._name = name
        self._arg_formatters = arg_formatters
        self.expression_format_string = expression_string
        self.latex_expression_format_string = latex_expression_string

        self.latex_name = latex_name

        self._description = None
        super(FunctionFormatter, self).__init__()

    @classmethod
    def _get_base_class(cls):
        return FunctionFormatter

    @classmethod
    def _get_object_type_name(cls):
        return 'function_formatter'

    def _get_format_kwargs(self, format_as_latex=False):
        """Create a dictionary containing argument name and format pairs.

        :param format_as_latex: If the format string is a latex formatted string.
        :return: Dictionary containing argument name and format pairs.
        :rtype: dict[str, str]
        """
        if format_as_latex:
            _par_name_string_dict = {_af.arg_name: _af.latex_name for _af in self.arg_formatters}
        else:
            _par_name_string_dict = {_af.arg_name: _af.name for _af in self.arg_formatters}
        return _par_name_string_dict

    def _get_formatted_name(self, format_as_latex=False):
        """Get the formatted function name.

        :param bool format_as_latex: If the format string is a latex formatted string.
        :return: The formatted function name.
        :rtype: str
        """
        if format_as_latex:
            return self._latex_name
        return self._name

    def _get_formatted_pars(self, with_par_values=True, n_significant_digits=2,
                            format_as_latex=False):
        """Get a list of the formatted parameters including their values. This can be turned off.

        :param bool with_par_values: If the strings should contain the parameter values.
        :param int n_significant_digits: The number of significant digits when using the parameter
            values.
        :param bool format_as_latex: If the string should be formatted as a latex string.
        :return: List of formatted parameter strings.
        :rtype: list[str]
        """
        if format_as_latex:
            _par_name_strings = [_pf.latex_name for _pf in self.par_formatters]
        else:
            _par_name_strings = [_pf.name for _pf in self.par_formatters]

        if with_par_values:
            _par_val_strings = []
            for _pf in self.par_formatters:
                _par_val_strings.append(_pf.get_formatted(with_name=False,
                                                          with_value=True,
                                                          with_errors=False,
                                                          n_significant_digits=n_significant_digits,
                                                          round_value_to_error=False,
                                                          format_as_latex=format_as_latex))
            return ["%s=%s" % (_pn, _pv) for _pn, _pv in zip(_par_name_strings, _par_val_strings)]
        return ["%s" % (_pn,) for _pn in _par_name_strings]

    def _get_formatted_expression(self, format_as_latex=False):
        """Get the formatted function expression.

        :param format_as_latex: If the string should be formatted as a latex string.
        :rtype: str
        """
        _kwargs = self._get_format_kwargs(format_as_latex=format_as_latex)
        if format_as_latex and self._latex_expr_string is not None:
            # use all arguments to format the expression string
            _par_expr_string = self._latex_expr_string.format(
                *[_af.latex_name for _af in self.arg_formatters], **_kwargs)
        elif not format_as_latex and self._expr_string is not None:
            _par_expr_string = self._expr_string.format(
                *[_af.name for _af in self.arg_formatters], **_kwargs)
        elif format_as_latex and self._latex_expr_string is None:
            _par_expr_string = self.DEFAULT_LATEX_EXPRESSION_STRING
        else:
            _par_expr_string = self.DEFAULT_EXPRESSION_STRING
        return _par_expr_string

    @property
    def expression_format_string(self):
        """A plain-text-formatted expression for the function.
        This function will replace all function parameters with their corresponding strings.
        For example the string "{a}*{x}+{b}" will turn into "A*x + B" when the name of the parameter
        a was set to "A", and the name of b is set to "B".

        :rtype: str
        """
        return self._expr_string

    @expression_format_string.setter
    def expression_format_string(self, expression_format_string):
        self._expr_string = expression_format_string
        try:
            self._get_formatted_expression(format_as_latex=False)
        except LookupError:  # fetch key and index errors. Other Errors have other causes.
            _af_arg_names = [_af.arg_name for _af in self.arg_formatters] if self.arg_formatters \
                else None
            raise ValueError("Expression string %s does not match argument structure %s"
                             % (expression_format_string, _af_arg_names))

    @property
    def latex_expression_format_string(self):
        r"""A LaTeX-formatted expression for the function.
        This function will replace all function parameters with their corresponding latex string.
        For example the string ``"{a}{x}+{b}"`` will turn into ``"A_0 x + B"`` when the latex name
        of the parameter ``a`` was set to ``"A_0"``, and the latex name of ``b`` is set to ``"B"``.

        .. note:
            Due to the way python handles string formatting, please always use two curly braces in
            standard LaTeX expressions. E.g. ``"\\frac{{{a}*{b}}}{{{x}-{c}}}"``.
            When not using a raw string, please double all backslashes as well! When using a raw
            string the above translates to ``r"\frac{{{a}*{b}}}{{{x}-{c}}}"``.

        :rtype: str
        """
        return self._latex_expr_string

    @latex_expression_format_string.setter
    def latex_expression_format_string(self, latex_expression_format_string):
        self._latex_expr_string = latex_expression_format_string
        try:
            self._get_formatted_expression(format_as_latex=True)
        except LookupError as _e:  # fetch key and index errors. Other Errors have other causes.
            _af_arg_names = [_af.arg_name for _af in self.arg_formatters] if self.arg_formatters \
                else None
            raise ValueError(
                "LaTeX expression string %s does not match argument structure %s" % (
                    latex_expression_format_string, _af_arg_names)
            ) from _e

    @property
    def name(self):
        """A plain-text-formatted string indicating the function name.

        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        # TODO: validate
        self._name = name

    @property
    def latex_name(self):
        """A LaTeX-formatted string indicating the function name.

        :rtype: str
        """
        return self._latex_name

    @latex_name.setter
    def latex_name(self, latex_name):
        if latex_name is None:
            self._latex_name = latexify_ascii(self.name)
        # TODO: validate
        else:
            self._latex_name = latex_name

    @property
    def description(self):
        """A short plain-text description of the function.

        :rtype: str
        """
        return self._description or "<no description provided>"

    @description.setter
    def description(self, description):
        # TODO: validate
        self._description = description

    @property
    def arg_formatters(self):
        """The list of :py:obj:`ParameterFormatter`-derived objects used for formatting all model
        function arguments.

        :rtype: list[ParameterFormatter]
        """
        return self._arg_formatters

    @property
    def par_formatters(self):
        """List of :py:obj:`ParameterFormatter`-derived objects used for formatting the fit parameters,
        excluding the independent parameter(s).

        :rtype: list[ParameterFormatter]
        """
        return self._arg_formatters  # same as all arguments for base, overwrite when necessary

    def get_formatted(self, with_par_values=False, n_significant_digits=2, format_as_latex=False,
                      with_expression=False):
        """Get a formatted string representing this model function.

        :param bool with_par_values: If :py:obj:`True`, output will include the value of each function parameter
                                     (e.g. ``f(a=1, b=2, ...)``).
        :param int n_significant_digits: Number of significant digits for rounding.
        :param bool format_as_latex: If :py:obj:`True`, the returned string will be formatted using LaTeX syntax.
        :param bool with_expression: If :py:obj:`True`, the returned string will include the expression assigned to the
                                     function.
        """
        _par_strings = self._get_formatted_pars(with_par_values=with_par_values,
                                                n_significant_digits=n_significant_digits,
                                                format_as_latex=format_as_latex)
        _par_expr_string = ""
        if with_expression:
            _par_expr_string = self._get_formatted_expression(format_as_latex=format_as_latex)

        if format_as_latex:
            _out_string = r"%s\left(%s\right)" % (self._latex_name, ", ".join(_par_strings))
            if _par_expr_string:
                _out_string += " = " + _par_expr_string
            _out_string = "$%s$" % (_out_string,)
        else:
            _out_string = "%s(%s)" % (self._name, ", ".join(_par_strings))
            if _par_expr_string:
                _out_string += " = " + _par_expr_string
        return _out_string


class CostFunctionFormatter(FunctionFormatter):
    """A Formatter class for Cost Functions.
    """

    def __init__(
            self, name, name_saturated=None, latex_name=None, latex_name_saturated=None,
            arg_formatters=None, expression_string=None, latex_expression_string=None):
        super(CostFunctionFormatter, self).__init__(
            name=name, latex_name=latex_name, arg_formatters=arg_formatters,
            expression_string=expression_string, latex_expression_string=latex_expression_string)
        self.name_saturated = name_saturated
        self.latex_name_saturated = latex_name_saturated

    @property
    def name_saturated(self):
        """A plain-text-formatted string indicating the saturated function name.

        :rtype: str
        """
        return self._name_saturated if self._name_saturated is not None else self._name

    @name_saturated.setter
    def name_saturated(self, name_saturated):
        # TODO: validate
        self._name_saturated = name_saturated

    @property
    def latex_name_saturated(self):
        """A LaTeX-formatted string indicating the saturated function name.

        :rtype: str
        """
        return self._latex_name_saturated if self._latex_name_saturated is not None \
            else self.latex_name

    @latex_name_saturated.setter
    def latex_name_saturated(self, latex_name_saturated):
        self._latex_name_saturated = latex_name_saturated  # TODO validate

    def get_formatted(
            self, value=None, n_degrees_of_freedom=None, with_name=True, saturated=False,
            with_value_per_ndf=True, format_as_latex=False):
        """Get a formatted string representing this cost function.

        :param value: Value of the cost function (if not :py:obj:`None`, the returned string will
            include this).
        :type value: float or None
        :param n_degrees_of_freedom: Number of degrees of freedom (if not :py:obj:`None`, the
            returned string will include this).
        :type n_degrees_of_freedom: int or None
        :param bool with_name: If :py:obj:`True`, the returned string will include the cost function
            name
        :param bool saturated: If :py:obj:`True`, the cost function name for the saturated
            Likelihood will be used (no effect for chi2).
        :param bool with_value_per_ndf: If :py:obj:`True`, the returned string will include the
            value-ndf ratio as a decimal value
        :param bool format_as_latex: If :py:obj:`True`, the returned string will be formatted using
            LaTeX syntax
        :rtype: str
        """

        if format_as_latex:
            _name = self.latex_name_saturated if saturated else self.latex_name
        else:
            _name = self.name_saturated if saturated else self.name
        _name_string = "%s" % _name
        _value_string = ""
        if value is not None:
            _value_string = "%.4g" % value
            if n_degrees_of_freedom is not None:
                if format_as_latex:
                    _name_string = r"%s / {\rm ndf}" % _name
                else:
                    _name_string = "%s / ndf" % _name
                _value_string = "%s / %d" % (_value_string, n_degrees_of_freedom)
                if with_value_per_ndf and n_degrees_of_freedom > 0:
                    _value_string = "%s = %.4g" % (_value_string,
                                                   float(value) / n_degrees_of_freedom)
        if format_as_latex:
            # replace scientific notation with power of ten (LaTeX only)
            _value_string = re.sub(r'(-?\d*\.?\d+?)0*e\+?(-?[0-9]*[1-9]?)',
                                     r'\1\\times10^{\2}', _value_string)

        if with_name:
            _out_string = "%s = %s" % (_name_string, _value_string)
        else:
            _out_string = "%s" % _value_string

        if format_as_latex:
            _out_string = "$%s$" % _out_string

        return _out_string


class ModelFunctionFormatter(FunctionFormatter):
    """A formatter class for model functions.

    This object stores the function name, formatted as a plain-text/LaTeX string, as well as a list
    of references to :py:obj:`ParameterFormatter` objects which contain information on how to format
    the model function arguments.
    Additionally formatting information about the independent variable is stored.

    Optionally, plain-text/LaTeX expression strings can be provided. These are strings representing
    the model function expression (i.e. mathematical formula).

    The formatted string is obtained by calling the :py:meth:`~.get_formatted` method.
    """

    @classmethod
    def _get_object_type_name(cls):
        return 'model_function_formatter'

    @property
    def par_formatters(self):
        # copy so we don't modify the original formatters
        formatters = copy.copy(self.arg_formatters)
        formatters.pop(0)  # first formatter is independent var, delete it
        return formatters

    def get_formatted(self, with_par_values=False, n_significant_digits=2, format_as_latex=False,
                      with_expression=False):
        """Create a formatted string representing this model function.

        :param bool with_par_values: If :py:obj:`True`, output will include the value of each
            function parameter (e.g. ``f(a=1, b=2, ...)``).
        :param int n_significant_digits: number of significant digits for rounding
        :param bool format_as_latex: If :py:obj:`True`, the returned string will be formatted using
            LaTeX syntax.
        :param bool with_expression: If :py:obj:`True`, the returned string will include the
            expression assigned to the function.
        :returns: The formatted string representing this model function.
        :rtype: str
        """
        _par_strings = self._get_formatted_pars(with_par_values=with_par_values,
                                                n_significant_digits=n_significant_digits,
                                                format_as_latex=format_as_latex)
        _par_expr_string = ""
        if with_expression:
            _par_expr_string = self._get_formatted_expression(format_as_latex=format_as_latex)

        x_formatter = self.arg_formatters[0]
        if format_as_latex:
            _out_string = r"%s\left(%s;%s\right)" % (
            self._latex_name, x_formatter.latex_name, ", ".join(_par_strings))
            if _par_expr_string:
                _out_string += " = " + _par_expr_string
            _out_string = "$%s$" % (_out_string,)
        else:
            _out_string = "%s(%s; %s)" % (self._name, x_formatter.name, ", ".join(_par_strings))
            if _par_expr_string:
                _out_string += " = " + _par_expr_string
        return _out_string



from ._base import ModelFunctionFormatterDReprBase, ParameterFormatterDReprBase
from .. import _AVAILABLE_REPRESENTATIONS
from .._yaml_base import YamlReaderMixin, YamlWriterMixin
from ..._base import ModelFunctionFormatter, ParameterFormatter
from ...indexed import IndexedModelFunctionFormatter

__all__ = ["ModelFunctionFormatterYamlWriter", "ModelFunctionFormatterYamlReader",
           "ParameterFormatterYamlWriter", "ParameterFormatterYamlReader"]


class ModelFunctionFormatterYamlWriter(YamlWriterMixin, ModelFunctionFormatterDReprBase):

    def __init__(self, model_function_formatter, output_io_handle):
        super(ModelFunctionFormatterYamlWriter, self).__init__(
            output_io_handle=output_io_handle,
            model_function_formatter=model_function_formatter)

    @classmethod
    def _make_representation(cls, model_function_formatter):
        """Create a representation of a :py:obj:`ModelFunctionFormatter` object as a dictionary.

        :param model_function_formatter: The :py:obj:`ModelFunctionFormatter` object to represent.
        :type model_function_formatter: ModelFunctionFormatter | IndexedModelFunctionFormatter
        :return: Dictionary containing all information about the :py:obj:`ModelFunctionFormatter`
            object.
        """
        _yaml_doc = dict()
        _class = model_function_formatter.__class__

        _type = cls._CLASS_TO_OBJECT_TYPE_NAME.get(_class, None)
        if _type is None:
            raise TypeError("Model function formatter unknown or not supported: %s" % _class)
        _yaml_doc['type'] = _type

        if _class is IndexedModelFunctionFormatter:
            _yaml_doc['index_name'] = model_function_formatter.index_name
            _yaml_doc['latex_index_name'] = model_function_formatter.latex_index_name
        elif _class is ModelFunctionFormatter:
            pass
        else:
            raise TypeError("Unknown formatter type!")

        _yaml_doc['name'] = model_function_formatter.name
        _yaml_doc['latex_name'] = model_function_formatter.latex_name

        _yaml_doc['expression_string'] = model_function_formatter.expression_format_string
        _yaml_doc['latex_expression_string'] = \
            model_function_formatter.latex_expression_format_string

        _arg_formatters_dict = dict()
        for _arg_formatter in model_function_formatter.arg_formatters:
            _arg_formatters_dict[_arg_formatter.name] = _arg_formatter.latex_name
        _yaml_doc['arg_formatters'] = _arg_formatters_dict

        # This is needed when saving a formatter without a model function object in order to
        # know the correct order of the arguments. This is because writing and then reading a yaml
        # file in Py2 will change the order of a dict or OrderedDict. This is not necessary for Py3!
        _yaml_doc['signature'] = [_arg_formatter.name for _arg_formatter in
                                  model_function_formatter.arg_formatters]
        return _yaml_doc


class ModelFunctionFormatterYamlReader(YamlReaderMixin, ModelFunctionFormatterDReprBase):

    def __init__(self, input_io_handle):
        super(ModelFunctionFormatterYamlReader, self).__init__(
            input_io_handle=input_io_handle,
            model_function_formatter=None)

    @classmethod
    def _modify_yaml_doc(cls, yaml_doc, kafe_object_class, name=None, signature=None):
        # only update keys in yaml doc if they don't exist
        # needed for setting name and signature when creating from a parametric model
        if 'name' not in yaml_doc and name is not None:
            yaml_doc['name'] = name
        if 'signature' not in yaml_doc and signature is not None:
            yaml_doc['signature'] = signature
        return yaml_doc

    @classmethod
    def _get_required_keywords(cls, yaml_doc, formatter_class):
        return ['name', 'signature']

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        # -- determine model function formatter class (only indexed and base)
        _type = yaml_doc.pop('type')
        _class = cls._OBJECT_TYPE_NAME_TO_CLASS.get(_type)

        _kwarg_list = ['name', 'latex_name', 'expression_string', 'latex_expression_string']
        if _class is IndexedModelFunctionFormatter:
            _kwarg_list.append('index_name')
            _kwarg_list.append('latex_index_name')
        elif _class is ModelFunctionFormatter:
            pass
        else:
            raise TypeError("Unknown formatter type!")
        _constructor_kwargs = {key: yaml_doc.pop(key, None) for key in _kwarg_list}

        _signature = yaml_doc.pop('signature')  # either list or dict, definitely a dict when
        # this is leftover from creating a model function
        _arg_formatters_yaml = yaml_doc.pop('arg_formatters', {})
        _arg_formatters = []
        for arg_name in _signature:
            _representation = _arg_formatters_yaml.pop(arg_name, {})
            if isinstance(_representation, str):
                _representation = {'latex_name': _representation}
            _representation.update({'id': arg_name})
            _arg_formatters.append(ParameterFormatterYamlReader._make_object(_representation))
        _constructor_kwargs['arg_formatters'] = _arg_formatters

        _model_function_formatter_object = _class(**_constructor_kwargs)

        return _model_function_formatter_object, yaml_doc


class ParameterFormatterYamlWriter(YamlWriterMixin, ParameterFormatterDReprBase):

    def __init__(self, model_parameter_formatter, output_io_handle):
        super(ParameterFormatterYamlWriter, self).__init__(
            output_io_handle=output_io_handle,
            model_parameter_formatter=model_parameter_formatter)

    @classmethod
    def _make_representation(cls, model_parameter_formatter):
        _yaml_doc = dict()
        _yaml_doc['id'] = model_parameter_formatter.arg_name
        _yaml_doc['name'] = model_parameter_formatter.name
        # parameter value and error are not part of the representation
        # because they belong to the parametric model
        _yaml_doc['latex_name'] = model_parameter_formatter.latex_name

        return _yaml_doc


class ParameterFormatterYamlReader(YamlReaderMixin, ParameterFormatterDReprBase):

    def __init__(self, input_io_handle):
        super(ParameterFormatterYamlReader, self).__init__(
            input_io_handle=input_io_handle,
            model_parameter_formatter=None)

    @classmethod
    def _type_required(cls):
        return False

    @classmethod
    def _get_required_keywords(cls, yaml_doc, kafe_object_class):
        return ['id']

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        # value and error are not part of the representation
        # because they belong to the parametric model
        _id = yaml_doc.pop('id')
        _name = yaml_doc.pop('name', None)
        _latex_name = yaml_doc.pop('latex_name', None)
        _model_parameter_formatter_object = ParameterFormatter(_id, name=_name,
                                                               latex_name=_latex_name)
        return _model_parameter_formatter_object, yaml_doc


# register the above classes in the module-level dictionary
ModelFunctionFormatterYamlReader._register_class(_AVAILABLE_REPRESENTATIONS)
ModelFunctionFormatterYamlWriter._register_class(_AVAILABLE_REPRESENTATIONS)
ParameterFormatterYamlReader._register_class(_AVAILABLE_REPRESENTATIONS)
ParameterFormatterYamlWriter._register_class(_AVAILABLE_REPRESENTATIONS)



import sys
import warnings
from collections import OrderedDict
from copy import copy

import numpy as np

from .cost import SharedCostFunction, MultiCostFunction
from .._base import FitBase
from ...core.fitters import NexusFitter
from ...core.error import SimpleGaussianError, MatrixGaussianError, CovMat
from ...core.fitters.nexus import Alias, Function, Array, Parameter
from ...tools import random_alphanumeric
from ..util import cholesky_decomposition, log_determinant, invert_matrix

__all__ = ['MultiFit']


class MultiFit(FitBase):
    """A MultiFit combines several regular fits into a combined fit object.
    Calling do_fit on the MultiFit will result in a numerical minimization of the sum of the cost functions of the
    individual fits.
    Parameters with the same name will be unified: their value will be the same across all individual fits.
    """

    _AXES = ()
    _MODEL_ERROR_NODE_NAMES = []

    def __init__(
            self, fit_list, minimizer=None, minimizer_kwargs=None,
            dynamic_error_algorithm="nonlinear"):
        """
        :param fit_list: List or Iterable of the individual fits from which to create the MultiFit.
        :type fit_list: collections.Iterable[FitBase]
        :param minimizer: The minimizer to use for fitting. Either ``'iminuit'``, ``'tminuit'``, ``'scipy'`` or
            ``None``.
        :type minimizer: str or None
        :param minimizer_kwargs: Dictionary with kwargs for the minimizer.
        :type minimizer_kwargs: dict

        :raises TypeError: If **fit_list** is not iterable.
        """
        # Cast Iterable to List, so that indexing works. Indexing is needed for some methods.
        self._fits = list(fit_list)  # will raise TypeError if fit_list is not iterable
        self._shared_error_dicts = dict()
        self._shared_error_nodes_initialized = False
        self._min_x_error = None
        super(MultiFit, self).__init__(
            data=None, model_function=None, cost_function=None, minimizer=minimizer,
            minimizer_kwargs=minimizer_kwargs, dynamic_error_algorithm=dynamic_error_algorithm)

    # -- private methods

    def _get_model_function_argument_formatters(self):
        _included_argument_names = set()
        _combined_argument_formatters = []
        for _fit in self._fits:
            for _argument_formatter in _fit._get_model_function_argument_formatters():
                if _argument_formatter.name not in _included_argument_names:
                    _combined_argument_formatters.append(_argument_formatter)
                    _included_argument_names.add(_argument_formatter.name)
        return _combined_argument_formatters

    def _get_model_function_parameter_formatters(self):
        _included_parameter_names = set()
        _combined_parameter_formatters = []
        for _fit in self._fits:
            for _parameter_formatter in _fit._get_model_function_parameter_formatters():
                if _parameter_formatter.name not in _included_parameter_names:
                    _combined_parameter_formatters.append(_parameter_formatter)
                    _included_parameter_names.add(_parameter_formatter.name)
        return _combined_parameter_formatters

    def _invalidate_total_error_cache(self):
        pass

    def _mark_errors_for_update(self):
        pass

    def _get_parameter_indices(self, singular_fit):
        return [self.parameter_names.index(_parameter_name) for _parameter_name in singular_fit.parameter_names]

    def _update_parameter_formatters(self, update_asymmetric_errors=False):
        if update_asymmetric_errors:
            _ = self.asymmetric_parameter_errors  # Ensure asymm par errs are calculated.
        for _fit in self._fits:
            _fit._update_parameter_formatters(update_asymmetric_errors=update_asymmetric_errors)

    def _update_singular_fits(self):
        _parameter_name_value_dict = self.parameter_name_value_dict
        for _fit in self._fits:
            _parameter_indices = self._get_parameter_indices(singular_fit=_fit)
            _asymmetric_parameter_errors = self._fitter.asymmetric_fit_parameter_errors_if_calculated
            if _asymmetric_parameter_errors is not None:
                _asymmetric_parameter_errors = _asymmetric_parameter_errors[_parameter_indices]
            _par_cor_mat = self.parameter_cor_mat
            if _par_cor_mat is not None:
                _par_cor_mat = _par_cor_mat[_parameter_indices][:, _parameter_indices]
            _par_cov_mat = self.parameter_cov_mat
            if _par_cov_mat is not None:
                _par_cov_mat = _par_cov_mat[_parameter_indices][:, _parameter_indices]
            _fit._loaded_result_dict = dict(
                did_fit=self.did_fit,
                parameter_errors=self.parameter_errors[_parameter_indices],
                parameter_cor_mat=_par_cor_mat,
                parameter_cov_mat=_par_cov_mat,
                asymmetric_parameter_errors=_asymmetric_parameter_errors
            )
        self._update_parameter_formatters()

    def _init_nexus(self):
        super(MultiFit, self)._init_nexus()

        self._combined_parameter_node_dict = OrderedDict()
        for _i, _fit_i in enumerate(self._fits):
            _original_cost_i = _fit_i._nexus.get('cost')
            _cost_alias_name_i = 'cost%s' % _i
            _cost_alias_i = Alias(ref=_original_cost_i, name=_cost_alias_name_i)
            self._nexus.add(_cost_alias_i, add_children=False)
            for _par_node in _fit_i.parameter_names:
                self._combined_parameter_node_dict[_par_node] = _fit_i._nexus.get(_par_node)
        for _par_node in self._combined_parameter_node_dict.values():
            self._nexus.add(_par_node)
            for _fit in self._fits:
                if _par_node.name in _fit.parameter_names:
                    _fit._nexus.add(node=_par_node, existing_behavior='replace')
        self._nexus.add(
            Array(nodes=self._combined_parameter_node_dict.values(), name='parameter_values'),
            existing_behavior="replace"
        )

        _log_det_names = []
        for _i, _fit_i in enumerate(self._fits):
            _fit_i._initialize_fitter()

            _x_data_node = _fit_i._nexus.get('x_data')
            if _x_data_node is not None:
                _x_data_name = 'x_data%s' % _i
                self._nexus.add(
                    Alias(ref=_x_data_node, name=_x_data_name),
                    add_children=False)

            _y_data_node = _fit_i._nexus.get('y_data')
            if _y_data_node is not None:
                _y_data_name = 'y_data%s' % _i
                self._nexus.add(
                    Alias(ref=_y_data_node, name=_y_data_name),
                    add_children=False)

            _log_det_node = _fit_i._nexus.get('total_cov_mat_log_determinant')
            if _log_det_node is not None:
                _log_det_name = 'total_cov_mat_log_determinant%s' % _i
                self._nexus.add(
                    Alias(ref=_log_det_node, name=_log_det_name),
                    add_children=False)
                _log_det_names.append(_log_det_name)

        self._nexus.add_function(
            lambda *log_dets: np.sum(log_dets),
            func_name='total_cov_mat_log_determinant', par_names=_log_det_names,
            existing_behavior="replace")

        _cost_functions = [_fit._cost_function for _fit in self._fits]
        _cost_names = ['cost%s' % _i for _i in range(len(self._fits))]
        self._cost_function = MultiCostFunction(
            singular_cost_functions=_cost_functions, cost_function_names=_cost_names)
        _cost_function_node = self._nexus.add_function(
            func=self._cost_function,
            func_name=self._cost_function.name,
            par_names=self._cost_function.arg_names
        )
        self._nexus.add_alias(
            name='cost', alias_for=_cost_function_node.name)

        self._initialize_fitter()

    def _init_shared_error_nodes(self):
        """
        initializes nexus nodes needed calculating cost with shared errors
        """
        from ..xy.fit import XYFit

        _data_indices = [0]  # Indices of edges of covariance block in combined matrix
        _fit_index_to_data_index = dict()  # Dict mapping fit index to lower data index index
        _cost_functions = []
        _cost_names = []
        _x_cov_mat_names = []
        _derivative_names = []
        _y_data_names = []
        _y_model_names = []
        _y_cov_mat_names = []
        for _i, _fit_i in enumerate(self._fits):
            if _fit_i._cost_function.is_chi2:
                _fit_index_to_data_index[_i] = len(_data_indices) - 1
                _data_indices.append(_data_indices[-1] + _fit_i.data_size)

                _x_cov_mat_name = 'x_cov_mat%s' % _i
                _derivatives_name = 'derivatives%s' % _i
                if isinstance(_fit_i, XYFit):
                    self._nexus.add(
                        Alias(ref=_fit_i._nexus.get('x_total_cov_mat'), name=_x_cov_mat_name),
                        add_children=False)

                    # Bind fit object (Python is call-by-value), otherwise all derivatives would
                    # use the same fit.
                    def _get_derivatives_func(fit):
                        return lambda: (fit._param_model.eval_model_function_derivative_by_x(
                            model_parameters=fit.parameter_values,
                            dx=0.01 * self._min_x_error
                        ) if self._min_x_error is not None else np.zeros(fit.data_size))

                    self._nexus.add(
                        Function(func=_get_derivatives_func(_fit_i), name=_derivatives_name),
                        add_children=False)
                    self._nexus.add_dependency(
                        name=_derivatives_name, depends_on='parameter_values')
                else:
                    self._nexus.add(Parameter(
                        np.zeros(_fit_i.data_size), name=_derivatives_name))
                    self._nexus.add(Parameter(
                        np.zeros((_fit_i.data_size, _fit_i.data_size)), name=_x_cov_mat_name))
                _x_cov_mat_names.append(_x_cov_mat_name)
                _derivative_names.append(_derivatives_name)

                _y_data_names.append('y_data%s' % _i)

                _y_model_name = 'y_model%s' % _i
                self._nexus.add(
                    Alias(ref=_fit_i._nexus.get('y_model'), name=_y_model_name),
                    add_children=False)
                _y_model_names.append(_y_model_name)

                _y_cov_mat_name = 'y_cov_mat%s' % _i
                self._nexus.add(
                    Alias(ref=_fit_i._nexus.get('y_total_cov_mat'), name=_y_cov_mat_name),
                    add_children=False)
                _y_cov_mat_names.append(_y_cov_mat_name)
            else:
                _cost_functions.append(_fit_i._cost_function)
                _cost_names.append('cost%s' % _i)

        # Combines 1-dimensional properties by concatenating them.
        def _combine_1d_property(*single_fit_properties):
            _combined_property = np.zeros(shape=_data_indices[-1])
            for _j, _single_fit_property in enumerate(single_fit_properties):
                _lower = _data_indices[_j]
                _upper = _data_indices[_j + 1]
                _combined_property[_lower:_upper] = _single_fit_property
            return _combined_property

        # Combines cov mats of single fits.
        # Shared errors are not added to the main diagonal blocks because they have already been
        # added to the individual fits.
        def _combine_cov_mats(axis_name, *single_fit_properties):
            _combined_property = np.zeros(shape=(_data_indices[-1], _data_indices[-1]))
            for _j, _single_fit_property in enumerate(single_fit_properties):
                _lower = _data_indices[_j]
                _upper = _data_indices[_j + 1]
                _combined_property[_lower:_upper, _lower:_upper] = _single_fit_property
            for _error_dict in self._shared_error_dicts.values():
                if _error_dict['axis'] != axis_name:
                    continue
                _error = _error_dict['err']
                for _j, _fit_index_j in enumerate(_error.fit_indices):
                    _data_index_j = _fit_index_to_data_index[_fit_index_j]
                    _lower_j = _data_indices[_data_index_j]
                    _upper_j = _data_indices[_data_index_j + 1]
                    for _k in range(_j):
                        _data_index_k = _fit_index_to_data_index[_error.fit_indices[_k]]
                        _lower_k = _data_indices[_data_index_k]
                        _upper_k = _data_indices[_data_index_k + 1]
                        _combined_property[_lower_j:_upper_j, _lower_k:_upper_k] += _error.cov_mat
                        _combined_property[_lower_k:_upper_k, _lower_j:_upper_j] += _error.cov_mat
            return _combined_property

        self._nexus.add_function(
            func=lambda *p: _combine_cov_mats('x', *p),
            func_name='x_cov_mat', par_names=_x_cov_mat_names, add_children=False)
        self._nexus.add_function(
            func=_combine_1d_property, func_name='derivatives', par_names=_derivative_names,
            add_children=False)

        self._nexus.add_function(
            func=_combine_1d_property, func_name='y_data',
            par_names=_y_data_names, add_children=False)
        self._nexus.add_alias(name='data', alias_for='y_data')
        self._nexus.add_function(
            func=_combine_1d_property, func_name='y_model',
            par_names=_y_model_names, add_children=False)
        self._nexus.add_alias(name='model', alias_for='y_model')
        self._nexus.add_function(
            func=lambda *p: _combine_cov_mats('y', *p),
            func_name='y_cov_mat', par_names=_y_cov_mat_names, add_children=False)

        def total_cov_mat_cholesky(x_cov_mat, derivatives, y_cov_mat):
            if self._min_x_error is not None:
                _cov_mat = y_cov_mat + x_cov_mat * np.outer(derivatives, derivatives)
            else:
                _cov_mat = y_cov_mat
            return cholesky_decomposition(_cov_mat)
        self._nexus.add_function(total_cov_mat_cholesky)
        self._nexus.add_function(log_determinant, func_name="total_cov_mat_log_determinant",
                                 par_names=["total_cov_mat_cholesky"], existing_behavior="replace")

        self._shared_cost_function = SharedCostFunction()
        self._nexus.add_function(
            func=self._shared_cost_function, func_name=self._shared_cost_function.name,
            par_names=self._shared_cost_function.arg_names
        )
        _cost_functions.append(self._shared_cost_function)
        _cost_names.append(self._shared_cost_function.name)
        self._cost_function = MultiCostFunction(
            singular_cost_functions=_cost_functions, cost_function_names=_cost_names)
        self._nexus.add_function(
            func=self._cost_function,
            func_name=self._cost_function.name,
            par_names=self._cost_function.arg_names,
            existing_behavior='replace'
        )
        self._initialize_fitter()

    def _initialize_fitter(self):
        self._fitter = NexusFitter(
            nexus=self._nexus, parameters_to_fit=list(self._combined_parameter_node_dict.keys()),
            parameter_to_minimize=self._cost_function.name, minimizer=self._minimizer,
            minimizer_kwargs=self._minimizer_kwargs
        )

    def _add_error_object(self, error_object, reference, name=None, axis=None):
        from ..indexed import IndexedFit
        from ..xy import XYFit

        if axis not in [None, 'x', 'y']:
            raise ValueError("axis must be one of: None, 'x', 'y'")
        _data_size_0 = self._fits[error_object.fit_indices[0]].data_size
        for _fit_index in error_object.fit_indices:
            if not self._fits[_fit_index]._cost_function.is_chi2:
                raise ValueError(
                    "Cannot add shared error because cost function of fit %s is not chi2!"
                    % _fit_index)
            if isinstance(self._fits[_fit_index], XYFit) and axis is None:
                raise ValueError("axis=None is ambiguous for fit %s because it is an XYFit!" % axis)
            if isinstance(self._fits[_fit_index], IndexedFit) and axis == 'x':
                raise ValueError(
                    "axis='x' is incompatible with fit %s because it is an IndexedFit!"
                    % _fit_index)
            if self._fits[_fit_index].data_size != _data_size_0:
                raise ValueError(
                    "Fit %s data_size not the same as data_size of Fit 0!" % _fit_index)

        if error_object.relative:
            if reference not in ('data', 'model'):
                raise ValueError(
                    "Error reference must be either 'model' or 'data' but received %s" % reference)
            if axis == 'x' and reference == 'data':
                _node_name = 'x_data%s'
            elif axis == 'y' and reference == 'data':
                _node_name = 'y_data%s'
            elif axis == 'x' and reference == 'model':
                _node_name = 'x_model%s'
                raise ValueError("Shared errors relative to model are not supported.")
            elif axis == 'y' and reference == 'model':
                _node_name = 'y_model%s'
                raise ValueError("Shared errors relative to model are not supported.")
            else:
                raise AssertionError()
            _reference_value = None
            _first_index = None
            for _fit_index in error_object.fit_indices:
                if _reference_value is None:
                    _reference_value = self._nexus.get(_node_name % _fit_index).value
                    _first_index = _fit_index
                else:
                    _other_reference_value = self._nexus.get(_node_name % _fit_index).value
                    if _reference_value.shape != _other_reference_value.shape or np.any(
                            (_reference_value - _other_reference_value) != 0):
                        raise ValueError(
                            "Cannot add a relative error for fits %s and %s because they have "
                            "conflicting references." % (_first_index, _fit_index))
            error_object.reference = _reference_value

        if axis is None:
            axis = 'y'
        if name in self._shared_error_dicts:
            raise ValueError("Error with name=%s already exists!" % name)
        if name is None:
            _name_is_unique = False
            while not _name_is_unique:
                name = random_alphanumeric(8)
                if name in self._shared_error_dicts:
                    continue
                for _fit_index in error_object.fit_indices:
                    _fit = self._fits[_fit_index]
                    if name in _fit.data_container._error_dicts:
                        continue
                    if name in _fit._param_model._error_dicts:
                        continue
                _name_is_unique = True

        _error_dict = dict(err=error_object, enabled=True, axis=axis, reference_name=reference)
        self._shared_error_dicts[name] = _error_dict
        for _fit_index in error_object.fit_indices:
            _fit = self._fits[_fit_index]
            if reference == "data":
                _target = _fit.data_container
            elif reference == "model":
                _target = _fit._param_model
            else:
                raise ValueError()
            _target._add_error_object(name=name, error_object=error_object, axis=axis)
        self._on_error_change()
        return name

    def _on_error_change(self):
        if not self._shared_error_nodes_initialized:
            self._init_shared_error_nodes()
            self._shared_error_nodes_initialized = True
        for _fit in self._fits:
            _fit._on_error_change()

        _x_errors = np.sqrt(np.diag(self._nexus.get("x_cov_mat").value))
        _non_zero_x_errors = _x_errors[_x_errors > 0.0]
        self._min_x_error = None if len(_non_zero_x_errors) == 0 else np.min(_non_zero_x_errors)

    def _set_new_data(self, new_data):
        raise NotImplementedError()

    def _set_new_parametric_model(self):
        raise NotImplementedError()

    def _set_data_as_model_ref(self):
        for _fit in self._fits:
            _fit._set_data_as_model_ref()

    def _iterative_fits_needed(self):
        for _fit in self._fits:
            if _fit._iterative_fits_needed():
                return True
        return False

    def _second_fit_needed(self):
        for _fit in self._fits:
            if _fit._second_fit_needed():
                return True
        return False

    def _pre_fit_iteration(self, first_fit=False):
        for _fit in self._fits:
            _fit._pre_fit_iteration(first_fit)

    def _post_fit_iteration(self, first_fit=False):
        for _fit in self._fits:
            _fit._post_fit_iteration(first_fit)

    # -- public properties

    @property
    def data(self):
        """List of the data of the individual fits."""
        return [_fit.data for _fit in self._fits]

    @data.setter
    def data(self, new_data):
        raise NotImplementedError("Use data setters for individual fits instead.")

    @property
    def data_error(self):
        raise NotImplementedError()

    @property
    def data_cov_mat(self):
        raise NotImplementedError()

    @property
    def data_cov_mat_inverse(self):
        raise NotImplementedError()

    @property
    def data_cor_mat(self):
        raise NotImplementedError()

    @property
    def model_error(self):
        raise NotImplementedError()

    @property
    def model_cov_mat(self):
        raise NotImplementedError()

    @property
    def model_cov_mat_inverse(self):
        raise NotImplementedError()

    @property
    def model_cor_mat(self):
        raise NotImplementedError()

    @property
    def total_error(self):
        _total_cov_mat = self.total_cov_mat
        if _total_cov_mat is None:
            return None
        else:
            return np.sqrt(np.diag(self.total_cov_mat))

    @property
    def total_cov_mat(self):
        if self._shared_error_dicts:
            _y_cov_mat = self._nexus.get("y_cov_mat").value
            _x_cov_mat = self._nexus.get("x_cov_mat").value
            _derivatives = self._nexus.get("derivatives").value
            return _y_cov_mat + _x_cov_mat * np.outer(_derivatives, _derivatives)
        else:
            _data_size = self.data_size
            if _data_size is None:
                return None
            _total_cov_mat = np.zeros((_data_size,) * 2)
            _lower = 0
            for _fit in self._fits:
                _upper = _lower + _fit.data_size
                _total_cov_mat[_lower:_upper, _lower:_upper] = _fit.total_cov_mat
                _lower = _upper
            return _total_cov_mat

    @property
    def data_container(self):
        """List of the data containers of the individual fits."""
        return [_fit.data_container for _fit in self._fits]

    @property
    def data_size(self):
        """Combined size of the data containers of the individual fits.

        :rtype: int
        """
        _data_sizes = [_fit.data_size for _fit in self._fits]
        if None in _data_sizes:
            return None
        else:
            return np.sum(_data_sizes)

    @property
    def has_errors(self):
        """``True`` if at least one uncertainty source is defined for any of the individual fits.

        :rtype: bool
        """
        return np.any([_fit.has_errors for _fit in self._fits])

    @property
    def has_data_errors(self):
        """``True`` if at least one uncertainty source is defined for the data of any of the individual fits.

        :rtype: bool
        """
        return np.any([_fit.has_data_errors for _fit in self._fits])

    @property
    def has_model_errors(self):
        """``True`` if at least one uncertainty source is defined for the model of any of the individual fits.

        :rtype: bool
        """
        return np.any([_fit.has_model_errors for _fit in self._fits])

    @property
    def model(self):
        """List of the model values of the individual fits."""
        return [_fit.model for _fit in self._fits]

    @property
    def model_count(self):
        """The number of model functions contained in the multifit. In most cases one for each singular fit.

        :rtype: int
        """
        return np.sum([_fit.model_count for _fit in self._fits])

    @property
    def model_label(self):
        """List of model labels for each singular fit."""
        return [_fit.model_label for _fit in self._fits]

    @model_label.setter
    def model_label(self, label):
        raise NotImplementedError("Use model_label setters for individual fits instead.")

    @property
    def fits(self):
        """List of individual fits on which the MultiFit is based on."""
        return copy(self._fits)  # shallow copy

    @property
    def asymmetric_parameter_errors(self):
        """The current asymmetric parameter uncertainties."""
        _asymm_par_errs = super(MultiFit, self).asymmetric_parameter_errors
        self._update_singular_fits()
        return _asymm_par_errs

    @property
    def ndf(self):
        _data_size = self.data_size
        if _data_size is None:
            return None
        else:
            return self.data_size - len(self._combined_parameter_node_dict.keys()) \
                + len(self._fitter.fixed_parameters)

    @property
    def goodness_of_fit(self):
        _gof_sum = 0.0
        for _fit in self._fits:
            if self._shared_error_nodes_initialized and _fit._cost_function.is_chi2:
                continue
            _gof = _fit.goodness_of_fit
            if _gof is None:
                return None
            _gof_sum += _gof
        if self._shared_error_nodes_initialized:
            _gof_sum += self._shared_cost_function.goodness_of_fit(
                *[self._nexus.get(_node_name).value
                  for _node_name in self._shared_cost_function.arg_names])
        return _gof_sum

    @property
    def chi2_probability(self):
        """The chi2 probability for the current model values."""
        _cost = self.cost_function_value
        if self._shared_error_nodes_initialized:
            _cost -= self._nexus.get("total_cov_mat_log_determinant").value
        for _fit in self._fits:
            _cost_func = _fit._cost_function
            if _cost_func.add_determinant_cost and not (
                    self._shared_error_nodes_initialized and _cost_func.is_chi2):
                _cost -= _fit._nexus.get("total_cov_mat_log_determinant").value
        return self._cost_function.chi2_probability(_cost, self.ndf)

    # -- public methods

    def add_matrix_error(self, err_matrix, matrix_type, fits, axis=None, name=None, err_val=None,
                         relative=False, reference='data', **kwargs):
        """Add a matrix uncertainty source for use in the fit.

        :param err_matrix: Covariance or correlation matrix.
        :param matrix_type: One of ``'covariance'``/``'cov'`` or ``'correlation'``/``'cor'``.
        :type matrix_type: str
        :param fits: The indices of the fits to add the error. If ``'all'``, the error is added to all fits.
        :type fits: int or collections.Iterable[int] or str
        :param axis: Axis of the individual fits to add the error to. ``y``/``1`` errors are treated as regular errors
                     for :py:class:`~kafe2.fit.IndexedFit`.
        :type axis: int or str
        :param name: Unique name for this uncertainty source. If ``None``, the name of the error source will be set to a
                     random alphanumeric string.
        :type name: str or None
        :param err_val: The pointwise uncertainties (mandatory if only a correlation matrix is given).
        :type err_val: collections.Iterable[float]
        :param relative: If ``True``, the covariance matrix and/or **err_val** will be interpreted as a *relative*
                         uncertainty.
        :type relative: bool
        :param reference: Either ``'data'`` or ``'model'``. Specifies which reference values to use when calculating
                          absolute errors from relative errors.
        :type reference: str
        :return: An error id which uniquely identifies the created error source.
        :rtype: str
        """
        # TODO relative errors
        if isinstance(fits, int):
            self._fits[fits].add_matrix_error(err_matrix=err_matrix, matrix_type=matrix_type, axis=axis, name=name,
                                              err_val=err_val, relative=relative, reference=reference, **kwargs)
        else:
            if fits == 'all':
                fits = list(range(len(self._fits)))
            _matrix_error = MatrixGaussianError(
                err_matrix=err_matrix, matrix_type=matrix_type, err_val=err_val, relative=relative,
                fit_indices=fits
            )
            _matrix_error.check_cov_mat_symmetry()
            return self._add_error_object(error_object=_matrix_error, reference=reference,
                                          name=name, axis=axis)

    def add_error(self, err_val, fits, axis=None, name=None, correlation=0, relative=False, reference='data', **kwargs):
        """Add an uncertainty source to the fit.

        :param err_val: Pointwise uncertainty/uncertainties for all data points.
        :type err_val: float or collections.Iterable[float]
        :param fits: The indices of the fits to add the error. If ``'all'``, the error is added to all fits.
        :type fits: int or collections.Iterable[int] or str
        :param axis: Axis of the individual fits to add the error to. ``y``/``1`` errors are treated as regular errors
                     for :py:class:`~kafe2.fit.IndexedFit`.
        :type axis: int or str
        :param name: Unique name for this uncertainty source. If ``None``, the name of the error source will be set to a
                     random alphanumeric string.
        :type name: str or None
        :param correlation: Correlation coefficient between any two distinct data points.
        :type correlation: float
        :param relative: If ``True``, **err_val** will be interpreted as a *relative* uncertainty.
        :type relative: bool
        :param reference: Either ``'data'`` or ``'model'``. Specifies which reference values to use when calculating
                          absolute errors from relative errors.
        :type reference: str
        :return: An error id which uniquely identifies the created error source.
        :rtype: str
        """
        if isinstance(fits, int):
            self._fits[fits].add_error(
                err_val=err_val, name=name, correlation=correlation, relative=relative,
                axis=axis, reference=reference, **kwargs
            )
        else:
            if fits == 'all':
                fits = list(range(len(self._fits)))
            try:
                err_val.ndim  # will raise if simple float
            except AttributeError:
                err_val = np.asarray(err_val, dtype=float)

            if err_val.ndim == 0:  # if dimensionless numpy array (i.e. float64), add a dimension
                err_val = np.ones(self._fits[fits[0]].data_size) * err_val
                # Consistent data size is checked later.
            _simple_error = SimpleGaussianError(
                err_val=err_val, corr_coeff=correlation, relative=relative, fit_indices=fits
            )
            return self._add_error_object(error_object=_simple_error, reference=reference,
                                          name=name, axis=axis)

    def assign_model_function_expression(self, expression_format_string, fit_index=None):
        """Assign a plain-text-formatted expression string to the model function of one of the individual fits.

        :param expression_format_string: The model function expression.
        :type expression_format_string: str
        :param fit_index: The index specifying the singular fit. If ``None``, all model functions will be changed.
        :type fit_index: int or None
        """
        if fit_index is None:
            for _fit in self._fits:
                _fit.assign_model_function_expression(
                    expression_format_string=expression_format_string)
        else:
            self._fits[fit_index].assign_model_function_expression(
                expression_format_string=expression_format_string)

    def assign_model_function_latex_name(self, latex_name, fit_index=None):
        """Assign a LaTeX-formatted string to be the model function name of one of the individual fits.

        :param latex_name: The LaTeX model function name.
        :type latex_name: str
        :param fit_index: The index specifying the singular fit. If ``None``, all model functions will be changed.
        :type fit_index: int or None
        """
        if fit_index is None:
            for _fit in self._fits:
                _fit.assign_model_function_latex_name(latex_name=latex_name)
        else:
            self._fits[fit_index].assign_model_function_latex_name(latex_name=latex_name)

    def assign_model_function_latex_expression(self, latex_expression_format_string, fit_index=None):
        """Assign a LaTeX-formatted expression string to the model function of one of the individual fits.
        Elements like ``'{par_name}'`` will be replaced automatically with the corresponding LaTeX names for the
        given parameter. These can be set with :py:meth:`~assign_parameter_latex_names`.

        :param latex_expression_format_string: The LaTeX model function expression.
        :type latex_expression_format_string: str
        :param fit_index: The index specifying the singular fit. If ``None``, all model functions will be changed.
        :type fit_index: int or None
        """
        if fit_index is None:
            for _fit in self._fits:
                _fit.assign_model_function_latex_expression(
                    latex_expression_format_string=latex_expression_format_string)
        else:
            self._fits[fit_index].assign_model_function_latex_expression(
                latex_expression_format_string=latex_expression_format_string)

    def assign_parameter_latex_names(self, **par_latex_names_dict):
        _keys = list(par_latex_names_dict.keys())
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Could not assign all latex names.*")
            for _fit in self._fits:
                _fit.assign_parameter_latex_names(**par_latex_names_dict)
                for _arg_formatter in _fit._get_model_function_argument_formatters():
                    try:
                        _keys.remove(_arg_formatter.name)
                    except ValueError:
                        pass
        if _keys:
            warnings.warn(
                "Could not assign all parameter latex names to single fits. Leftover: {}".format(
                    _keys))

    def disable_error(self, err_id):
        for _fit in self._fits:
            _fit.disable_error(err_id=err_id)

    def fix_parameter(self, name, value=None):
        self._fitter.fix_parameter(name=name, value=value)
        # get fixed value before setting it in the individual fits
        _val = self._fitter.fixed_parameters[name]
        for fit in self._fits:
            if name not in fit.parameter_names:
                continue  # skip if sub fit is not dependent on the given par
            fit.fix_parameter(name, _val)  # default values might not have been overwritten, use _val

    def release_parameter(self, name):
        self._fitter.release_parameter(name)
        for fit in self._fits:  # update formatters of individual fits
            if name not in fit.parameter_names:
                continue  # skip if sub fit is not dependent on the given par
            fit.release_parameter(name)

    def do_fit(self, asymmetric_parameter_errors=False):
        _fit_result = super(MultiFit, self).do_fit(
            asymmetric_parameter_errors=asymmetric_parameter_errors)
        self._update_singular_fits()
        return _fit_result

    def get_matching_errors(self, fit_index=None, matching_criteria=None, matching_type='equal'):
        """Return a list of uncertainty objects fulfilling the specified matching criteria.

        Valid keys for **matching_criteria**:
            * ``name`` (the unique error name)
            * ``type`` (either ``'simple'`` or ``'matrix'``)
            * ``correlated`` (bool, only matches simple errors!)
            * ``reference`` (either ``'model'`` or ``'data'``)

        .. note::
            The error objects contained in the dictionary are not copies, but the original error objects.
            Modifying them is possible, but not recommended.
            If you do modify any of them, the changes will not be reflected in the total error calculation until the
            error cache is cleared. This can be done by calling the private dataset method
            :py:meth:`~kafe2.fit._base.DataContainerBase._clear_total_error_cache`.

        :param fit_index: Index for which fit the method should be executed. If ``None`` the function will return the
                          matched errors for all singular fits.
        :type fit_index: int or None
        :param matching_criteria: Key-value pairs specifying matching criteria. The resulting error array will only
                                  contain error objects matching *all* provided criteria.
                                  If ``None``, all error objects are returned.
        :type matching_criteria: dict or None
        :param matching_type: How to perform the matching.
                              If ``'equal'``, the value in ``matching_criteria`` is checked for equality against the
                              stored value.
                              If ``'regex'``, the value in ``matching_criteria`` is interpreted as a regular expression
                              and is matched against the stored value.
        :type matching_type: str
        :return: Dict mapping error name to :py:obj:`~kafe2.core.error.GaussianErrorBase`-derived error objects.
        :rtype: dict[str, kafe2.core.error.GaussianErrorBase]
        """
        if fit_index is None:
            _combined_matching_errors = dict()
            for _fit in self._fits:
                _matching_errors = _fit.get_matching_errors(
                    matching_criteria=matching_criteria, matching_type=matching_type)
                for _key in _matching_errors:
                    # FATAL: there is an error with the same name in separate fits
                    assert _key not in _combined_matching_errors
                    _combined_matching_errors[_key] = _matching_errors[_key]
            return _combined_matching_errors
        return self._fits[fit_index].get_matching_errors(matching_criteria=matching_criteria,
                                                         matching_type=matching_type)

    def report(self, output_stream=sys.stdout, show_data=True, show_model=True, show_fit_results=True,
               asymmetric_parameter_errors=False):
        """Print a summary for each fit state and/or the multifit results.

        :param output_stream: The output stream to which the report should be printed.
        :type output_stream: io.TextIOBase
        :param show_data: If ``True``, print out information about the data for each fit.
        :type show_data: bool
        :param show_model: If ``True``, print out information about the parametric model for each fit.
        :type show_model: bool
        :param show_fit_results: If ``True``, print out information about the combined fit results.
        :type show_fit_results: bool
        :param asymmetric_parameter_errors: If ``True``, use two different parameter errors for up/down directions.
        :type asymmetric_parameter_errors: bool
        """
        _indent = ' ' * 4

        for _i, _fit in enumerate(self._fits):
            _header_string = '#########\n'
            if _i > 9:
                _header_string = '#' + _header_string
            if _i > 99:
                _header_string = '#' + _header_string
            output_stream.write(_header_string)
            output_stream.write('# Fit %s #\n' % _i)
            output_stream.write(_header_string)
            output_stream.write('\n')

            if show_data:
                _fit._report_data(output_stream=output_stream, indent=_indent, indentation_level=1)
            if show_model:
                _fit._report_model(output_stream=output_stream, indent=_indent, indentation_level=1)
        self._report_fit_results(output_stream=output_stream, indent=_indent, indentation_level=0,
                                 asymmetric_parameter_errors=asymmetric_parameter_errors)



import abc
import getpass
import datetime
import six


@six.add_metaclass(abc.ABCMeta)
class GenericDReprBase(object):
    BASE_OBJECT_TYPE_NAME = None
    DREPR_FLAVOR_NAME = None
    DREPR_ROLE_NAME = None

    @classmethod
    def _register_class(cls, global_dict):
        _registered_roles = global_dict.setdefault(cls.BASE_OBJECT_TYPE_NAME, dict())
        _registered_formats_for_role = _registered_roles.setdefault(cls.DREPR_ROLE_NAME, dict())
        _registered_formats_for_role[cls.DREPR_FLAVOR_NAME] = cls


class DReprWriterMixin(object):
    
    DREPR_ROLE_NAME = 'writer'

    """
    A "mixin" class for creating a data representation writer.
    Inheriting from this class in addition to a DRepr class for
    a particular object type adds methods for writing to
    an output stream.

    Derived classes should inherit from :py:class:`DReprWriterMixin` and the
    relevant ``DRepr`` class (in that order).
    """
    def __init__(self, output_io_handle, *args, **kwargs):
        """
        Mixin constructor: sets the output handle

        :param output_io_handle: handle for output stream or file
        :type output_io_handle: :py:class:`~kafe2.fit.io.IOStreamHandle`-derived
        """
        self._ohandle = output_io_handle
        self._representation = None
        super(DReprWriterMixin, self).__init__(*args, **kwargs)

    @staticmethod
    def _make_representation(kafe_object):
        """Implement this method for the different data formats"""
        pass

    def _get_preface_comment(self):
        return '# kafe2 %s %s representation written by %s on %s.\n' % (
            self._kafe_object.__class__.__name__,
            self.DREPR_FLAVOR_NAME,
            getpass.getuser(),
            datetime.datetime.now().strftime('%d.%m.%Y, %H:%M')
        )

    def write(self):
        self._representation = self._make_representation(self._kafe_object)
        with self._ohandle as _h:
            _h.write(self._get_preface_comment())
            _h.write(self._representation)


class DReprReaderMixin(object):
    
    DREPR_ROLE_NAME = 'reader'
    
    """
    A "mixin" class for creating a data representation writer.
    Inheriting from this class in addition to a DRepr class for
    a particular object type adds methods for reading from
    an input stream.

    Derived classes should inherit from :py:class:`DReprReaderMixin` and the
    relevant ``DRepr`` class (in that order).
    """
    def __init__(self, input_io_handle, *args, **kwargs):
        """
        Mixin constructor: sets the input handle

        :param input_io_handle: handle for input stream or file
        :type input_io_handle: :py:class:`~kafe2.fit.io.IOStreamHandle`-derived
        """
        self._ihandle = input_io_handle
        self._object = None
        super(DReprReaderMixin, self).__init__(*args, **kwargs)

    @staticmethod
    def _make_object(representation):
        """
        this method is supposed to create and return an object from the representation.
        It has to be implemented by subclasses.
        """
        pass

    def read(self):
        with self._ihandle as _h:
            self._representation = _h.read()
        return self._make_object(self._representation)



import textwrap
import tokenize

import numpy as np
import six

from ._base import ModelFunctionDReprBase, ParametricModelDReprBase
from .. import _AVAILABLE_REPRESENTATIONS
from .._yaml_base import YamlReaderMixin, YamlWriterMixin
from ..error import common_error_tools
from ..format import ModelFunctionFormatterYamlReader, ModelFunctionFormatterYamlWriter
from ..._base import ModelFunctionBase
from ....fit import HistModelFunction, HistParametricModel, IndexedModelFunction, \
    IndexedParametricModel, UnbinnedParametricModel, XYParametricModel
from ....fit.util import function_library

__all__ = ['ModelFunctionYamlWriter', 'ModelFunctionYamlReader', 'ParametricModelYamlWriter',
           'ParametricModelYamlReader']
KNOWN_MODEL_FUNCTIONS = (ModelFunctionBase, HistModelFunction, IndexedModelFunction)
KNOWN_PARAMETRIC_MODELS = (XYParametricModel, HistParametricModel, IndexedParametricModel,
                           UnbinnedParametricModel)


def _parse_function(input_string):
    """converts a string of python code into a python function object"""
    _tokens = tokenize.generate_tokens(six.StringIO(input_string).readline)
    for _toknum, _tokval, _spos, _epos, _line_string in _tokens:
        if _tokval in ModelFunctionYamlReader.FORBIDDEN_TOKENS:
            raise ValueError("Encountered forbidden token '%s' in user-entered code on line '%s'."
                             % (_tokval, _line_string))

    if "__" in input_string:
        raise ValueError("Model function input must not contain '__'!")

    _imports = ""
    _imports += "import numpy as np\n"  # import numpy
    # import scipy if installed
    try:
        import scipy
        _imports += "import scipy\n"
    except ImportError:
        pass

    __locals_pointer = [None, None]  # TODO better solution?
    # save locals before function definition
    _exec_string = _imports + "__locals_pointer[0] = __locals().copy()\n" + input_string
    # save locals after function definition
    _exec_string = _exec_string + "\n__locals_pointer[1] = __locals().copy()"
    _restricted_builtins = __builtins__.copy()
    for _forbidden_token in ModelFunctionYamlReader.FORBIDDEN_TOKENS:
        _restricted_builtins.pop(_forbidden_token, None)
    _restricted_builtins['__import__'] = __import__
    exec(_exec_string, {"__builtins__": _restricted_builtins, "__locals": locals,
                        "__locals_pointer": __locals_pointer})
    _locals_pre, _locals_post = __locals_pointer[0].values(), __locals_pointer[1].values()
    _new_references = []
    for _post_reference in _locals_post:
        if _post_reference not in _locals_pre:
            _new_references.append(_post_reference)
    if len(_new_references) != 1:
        raise ValueError(
            "Expected to receive exactly one new reference as a model function but instead received"
            "%s in the following string:\n%s" % (len(_new_references), input_string))
    return _new_references[0]


def _process_function_code_for_dump(source_code):
    source_code = textwrap.dedent(source_code)  # remove indentation
    source_code = source_code.replace("@staticmethod\n", "")  # remove @staticmethod decorator
    # TODO what about other decorators?
    return source_code


class ModelFunctionYamlWriter(YamlWriterMixin, ModelFunctionDReprBase):

    def __init__(self, model_function, output_io_handle):
        super(ModelFunctionYamlWriter, self).__init__(
            output_io_handle=output_io_handle,
            model_function=model_function)

    @classmethod
    def _make_representation(cls, model_function):
        _yaml_doc = dict()

        # -- determine model function type
        _class = model_function.__class__
        _type = cls._CLASS_TO_OBJECT_TYPE_NAME.get(_class, None)
        if _type is None:
            raise TypeError("Model function type unknown or not supported: %s" % _class)
        if _type != 'base':  # 'base' is default, no need to store type
            _yaml_doc['type'] = _type  # store all other custom types
        _yaml_doc['model_function_formatter'] =\
            ModelFunctionFormatterYamlWriter._make_representation(model_function.formatter)

        _yaml_doc['python_code'] = _process_function_code_for_dump(model_function.source_code)

        return _yaml_doc


class ModelFunctionYamlReader(YamlReaderMixin, ModelFunctionDReprBase):
    FORBIDDEN_TOKENS = ['compile', 'eval', 'exec', 'execfile', 'file', 'global', 'globals',
                        'import', '__import__',
                        'input', ' locals', 'nonlocal', 'open', 'reload', 'self', 'super']

    def __init__(self, input_io_handle):
        super(ModelFunctionYamlReader, self).__init__(
            input_io_handle=input_io_handle,
            model_function=None)

    @classmethod
    def _get_subspace_override_dict(cls, model_function_class):
        _override_dict = {'name': 'model_function_formatter',
                          'latex_name': 'model_function_formatter',
                          'arg_formatters': 'model_function_formatter',
                          'expression_string': 'model_function_formatter',
                          'latex_expression_string': 'model_function_formatter',
                          'signature': 'model_function_formatter'}
        if model_function_class is IndexedModelFunction:
            _override_dict['index_name'] = 'model_function_formatter'
            _override_dict['latex_index_name'] = 'model_function_formatter'
        return _override_dict

    @classmethod
    def _get_required_keywords(cls, yaml_doc, model_function_class):
        if issubclass(model_function_class, ModelFunctionBase):
            return ['python_code']
        raise TypeError("Unknown model function class: %s" % model_function_class)

    @classmethod
    def _process_string(cls, string_representation, default_type):
        return dict(type=default_type, python_code=string_representation)

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        """Converts the given yaml dictionary to a :py:obj:``ModelFunctionBase`` or derived object.

        :param yaml_doc: the yaml doc to convert
        :type yaml_doc: dict"""
        # -- determine model function class from type
        _model_function_type = yaml_doc.pop('type', 'base')  # if no type is specified assume base
        _class = cls._OBJECT_TYPE_NAME_TO_CLASS.get(_model_function_type)
        if _class not in KNOWN_MODEL_FUNCTIONS:
            raise TypeError("Unknown model function class: %s" % _class)

        _raw_string = yaml_doc.pop("python_code")
        try:
            _model_function_object = _class(_raw_string)
        except ValueError:
            _parsed_function = _parse_function(_raw_string)
            _model_function_object = _class(_parsed_function)
            _model_function_object._source_code = _raw_string

        # construct model function formatter if specified
        _model_function_formatter_yaml = yaml_doc.pop('model_function_formatter', None)
        if _model_function_formatter_yaml:
            _name = _model_function_object.name
            _signature = _model_function_object.signature.parameters
            _model_function_object._formatter = ModelFunctionFormatterYamlReader._make_object(
                _model_function_formatter_yaml, default_type=_model_function_type, name=_name,
                signature=_signature)

        return _model_function_object, yaml_doc


class ParametricModelYamlWriter(YamlWriterMixin, ParametricModelDReprBase):
    DUMPER = common_error_tools.MatrixYamlDumper

    def __init__(self, parametric_model, output_io_handle):
        super(ParametricModelYamlWriter, self).__init__(output_io_handle=output_io_handle,
                                                        parametric_model=parametric_model)

    @classmethod
    def _make_representation(cls, parametric_model):
        """Create a dictionary representation of a parametric model.

        :param parametric_model: The parametric model to convert.
        :type parametric_model: kafe2.fit.histogram.HistParametricModel | kafe2.fit.indexed.IndexedParametricModel |
                                kafe2.fit.xy.XYParametricModel
        """
        _yaml_doc = dict()

        # -- determine model function type
        _class = parametric_model.__class__
        _type = cls._CLASS_TO_OBJECT_TYPE_NAME.get(_class, None)
        if _type is None:
            raise TypeError("Parametric model type unknown or not supported: %s" % _class)
        _yaml_doc['type'] = _type

        # -- write representation for model types
        if _class is HistParametricModel:
            _yaml_doc['n_bins'] = int(parametric_model.n_bins)
            _yaml_doc['bin_range'] = list(map(float, parametric_model.bin_range))
            _yaml_doc['model_density_function'] = ModelFunctionYamlWriter._make_representation(
                parametric_model._model_function_object)
            _yaml_doc['bin_edges'] = list(map(float, parametric_model.bin_edges))
            if isinstance(parametric_model.bin_evaluation, str):
                _yaml_doc['bin_evaluation'] = parametric_model.bin_evaluation_string
            else:
                _yaml_doc['bin_evaluation'] = _process_function_code_for_dump(
                    parametric_model.bin_evaluation_string)
            _yaml_doc['density'] = parametric_model.density
        elif _class is IndexedParametricModel:
            _yaml_doc['shape_like'] = parametric_model.data.tolist()
            _yaml_doc['model_function'] = ModelFunctionYamlWriter._make_representation(
                parametric_model._model_function_object)
        elif _class is UnbinnedParametricModel:
            _yaml_doc['data'] = parametric_model.support.tolist()
            _yaml_doc['model_function'] = ModelFunctionYamlWriter._make_representation(
                parametric_model._model_function_object)
        elif _class is XYParametricModel:
            _yaml_doc['x_data'] = parametric_model.x.tolist()
            _yaml_doc['y_data'] = parametric_model.y.tolist()
            _yaml_doc['model_function'] = ModelFunctionYamlWriter._make_representation(
                parametric_model._model_function_object)
        else:
            raise TypeError("Unkonwn parametric model type")

        # convert all numpy array entries to regular float, then convert to list, improves readability
        _yaml_doc['model_parameters'] = np.array(parametric_model.parameters, dtype=float).tolist()

        # write model label for all types
        _yaml_doc['model_label'] = parametric_model.label

        # -- write error representation for all container types
        if parametric_model.has_errors:
            _yaml_doc = common_error_tools.write_errors_to_yaml(container=parametric_model,
                                                                yaml_doc=_yaml_doc)

        return _yaml_doc


class ParametricModelYamlReader(YamlReaderMixin, ParametricModelDReprBase):

    def __init__(self, input_io_handle):
        super(ParametricModelYamlReader, self).__init__(
            input_io_handle=input_io_handle,
            parametric_model=None)

    @classmethod
    def _get_subspace_override_dict(cls, parametric_model_class):
        _override_dict = {'arg_formatters': 'model_function',
                          'model_function_formatter': 'model_function'}

        if parametric_model_class is HistParametricModel:
            _override_dict['model_density_function_name'] = 'model_function'
            _override_dict['latex_model_density_function_name'] = 'model_function'
            _override_dict['expression_string'] = 'model_function'
            _override_dict['latex_expression_string'] = 'model_function'
        elif parametric_model_class is IndexedParametricModel:
            _override_dict['model_function_name'] = 'model_function'
            _override_dict['latex_model_function_name'] = 'model_function'
            _override_dict['index_name'] = 'model_function'
            _override_dict['latex_index_name'] = 'model_function'
            _override_dict['expression_string'] = 'model_function'
            _override_dict['latex_expression_string'] = 'model_function'
        elif parametric_model_class is XYParametricModel or parametric_model_class is UnbinnedParametricModel:
            _override_dict['model_function_name'] = 'model_function'
            _override_dict['latex_model_function_name'] = 'model_function'
            _override_dict['expression_string'] = 'model_function'
            _override_dict['latex_expression_string'] = 'model_function'
        else:
            raise TypeError("Unknown parametric model type")
        return _override_dict

    @classmethod
    def _get_required_keywords(cls, yaml_doc, parametric_model_class):
        if parametric_model_class is HistParametricModel:
            return []
        if parametric_model_class is UnbinnedParametricModel:
            return ['data']
        if parametric_model_class is IndexedParametricModel:
            return ['model_function']
        if parametric_model_class is XYParametricModel:
            return ['x_data']

    @classmethod
    def _get_ignored_if_none_keywords(cls):
        return ["model_density_func_antiderivative"]

    @classmethod
    def _modify_yaml_doc(cls, yaml_doc, kafe_object_class, dataset=None, **kwargs):
        if dataset:
            if kafe_object_class is HistParametricModel:
                if 'bin_edges' not in yaml_doc and ('n_bins' not in yaml_doc or 'bin_range' not
                                                    in yaml_doc):
                    yaml_doc['bin_edges'] = dataset.bin_edges
            elif kafe_object_class is IndexedParametricModel:
                if 'shape_like' not in yaml_doc:
                    yaml_doc['shape_like'] = dataset.data
            elif kafe_object_class is UnbinnedParametricModel:
                if 'data' not in yaml_doc:
                    yaml_doc['data'] = dataset.data
            elif kafe_object_class is XYParametricModel:
                if 'x_data' not in yaml_doc:
                    yaml_doc['x_data'] = dataset.x
        super(ParametricModelYamlReader, cls)._modify_yaml_doc(yaml_doc, kafe_object_class,
                                                               **kwargs)
        return yaml_doc

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        # -- determine model function class from type
        _parametric_model_type = yaml_doc.pop('type')
        _class = cls._OBJECT_TYPE_NAME_TO_CLASS.get(_parametric_model_type, None)

        # _kwarg_list = kwargs that directly correspond to unchanging scalar yaml entries
        _kwarg_list = []
        _constructor_kwargs = {}
        _hist_model_bin_evaluation_source = None  # Set if an antiderivative function is read in.
        if _class is HistParametricModel:
            if 'bin_edges' in yaml_doc:
                _bin_edges = yaml_doc.pop('bin_edges')
                # bin_edges overrides n_bins and bin_range
                yaml_doc.pop('n_bins', None)
                yaml_doc.pop('bin_range', None)
                _constructor_kwargs['bin_edges'] = _bin_edges
                _constructor_kwargs['n_bins'] = len(_bin_edges) - 1
                _constructor_kwargs['bin_range'] = [_bin_edges[0], _bin_edges[-1]]

            elif 'n_bins' in yaml_doc and 'bin_range' in yaml_doc:
                _kwarg_list.append('n_bins')
                _kwarg_list.append('bin_range')
            else:
                raise ValueError("When reading in a histogram parametric model either "
                                 "bin_edges or n_bins and bin_range have to be specified!")
            if 'bin_evaluation' in yaml_doc:
                _bin_evaluation = yaml_doc.pop('bin_evaluation')
                if 'def' in _bin_evaluation:
                    _constructor_kwargs['bin_evaluation'] = _parse_function(_bin_evaluation)
                    _hist_model_bin_evaluation_source = _bin_evaluation
                else:
                    _constructor_kwargs['bin_evaluation'] = _bin_evaluation
            if 'density' in yaml_doc:
                _constructor_kwargs['density'] = yaml_doc.pop('density')
        elif _class is IndexedParametricModel:
            _kwarg_list.append('shape_like')
        elif _class is UnbinnedParametricModel:
            _kwarg_list.append('data')
        elif _class is XYParametricModel:
            _kwarg_list.append('x_data')
            yaml_doc.pop('y_data', None)  # remove y_data from dict
        else:
            raise TypeError("Unknown parametric model type")

        _model_func = None
        if _class is HistParametricModel:
            _model_func_entry = yaml_doc.pop('model_density_function', None)
            if _model_func_entry:
                _model_func = ModelFunctionYamlReader._make_object(
                    _model_func_entry, default_type=_parametric_model_type)
                _constructor_kwargs['model_density_func'] = _model_func
        elif _class is UnbinnedParametricModel:
            _model_func_entry = yaml_doc.pop('model_function', None)
            if _model_func_entry:
                _model_func = ModelFunctionYamlReader._make_object(
                    _model_func_entry, default_type=_parametric_model_type)
                _constructor_kwargs['model_density_function'] = _model_func
        elif _class in (IndexedParametricModel, XYParametricModel):
            _model_func_entry = yaml_doc.pop('model_function', None)
            if _model_func_entry:
                _model_func = ModelFunctionYamlReader._make_object(
                    _model_func_entry, default_type=_parametric_model_type)
                _constructor_kwargs['model_func'] = _model_func
        else:
            raise TypeError('Unknown model type: %s' % _parametric_model_type)

        if _model_func:
            # if model parameters are given, apply those to the model function
            # if not use model function defaults            
            _given_parameters = yaml_doc.pop('model_parameters', None)
            if _given_parameters:
                _model_func.defaults = _given_parameters
                _constructor_kwargs['model_parameters'] = _given_parameters
            else:
                _constructor_kwargs['model_parameters'] = _model_func.defaults

        _constructor_kwargs.update({key: yaml_doc.pop(key, None) for key in _kwarg_list})
        _parametric_model_object = _class(**_constructor_kwargs)
        if _hist_model_bin_evaluation_source is not None:
            _parametric_model_object._bin_evaluation_string = _hist_model_bin_evaluation_source

        # add the label for all types
        _parametric_model_object.label = yaml_doc.pop('model_label', None)

        # -- process error sources
        _parametric_model_object, yaml_doc = common_error_tools.process_error_sources(
            container_obj=_parametric_model_object, yaml_doc=yaml_doc)

        return _parametric_model_object, yaml_doc


# register the above classes in the module-level dictionary
ModelFunctionYamlReader._register_class(_AVAILABLE_REPRESENTATIONS)
ModelFunctionYamlWriter._register_class(_AVAILABLE_REPRESENTATIONS)
ParametricModelYamlReader._register_class(_AVAILABLE_REPRESENTATIONS)
ParametricModelYamlWriter._register_class(_AVAILABLE_REPRESENTATIONS)



from ..indexed import IndexedContainer

__all__ = ['UnbinnedContainer']


class UnbinnedContainer(IndexedContainer):
    """
    This object is a specialized data container for series of measurements.

    """
    def __init__(self, data, dtype=float):
        """
        Construct a container for indexed data:

        :param data: a one-dimensional array of measurements
        :type data: iterable of type <dtype>
        :param dtype: data type of the measurements
        :type dtype: type
        """
        super(UnbinnedContainer, self).__init__(data, dtype)

    def add_error(self):
        raise NotImplementedError("Unbinned fits don't support errors")

    def add_matrix_error(self):
        raise NotImplementedError("Unbinned fits don't support errors")



import numpy as np

from ...core.error import MatrixGaussianError
from .._base import DataContainerBase


__all__ = ['IndexedContainer']


class IndexedContainer(DataContainerBase):
    """
    This object is a specialized data container for series of indexed measurements.

    """
    def __init__(self, data, dtype=float):
        """
        Construct a container for indexed data:

        :param data: a one-dimensional array of measurements
        :type data: iterable of type <dtype>
        :param dtype: data type of the measurements
        :type dtype: type
        """
        super(IndexedContainer, self).__init__()
        self._data = np.array(data, dtype=dtype)
        if self._data.ndim != 1:
            raise ValueError("Input data must be one-dimensional but received"
                             f" data with {self._data.ndim} dimensions.")

    # -- private methods

    def _calculate_total_error(self):
        _sz = self.size
        _tmp_cov_mat = np.zeros((_sz, _sz))
        for _err_dict in self._error_dicts.values():
            if not _err_dict['enabled']:
                continue
            _tmp_cov_mat += _err_dict['err'].cov_mat

        _total_err = MatrixGaussianError(_tmp_cov_mat, 'cov', relative=False, reference=self.data)
        self._total_error = _total_err

    def _clear_total_error_cache(self):
        self._total_error = None

    def _get_error_reference(self):
        return self._data

    # -- public properties

    @property
    def size(self):
        """number of data points"""
        return len(self._data)

    @property
    def data(self):
        """container data (one-dimensional :py:obj:`numpy.ndarray`)"""
        return self._data.copy()  # copy to ensure no modification by user

    @data.setter
    def data(self, data):
        _data = np.squeeze(np.array(data, dtype=float))
        if len(_data.shape) > 1:
            raise ValueError("IndexedContainer data must be 1-d array of floats! Got shape: %r..."
                             % (_data.shape,))
        self._data[:] = _data
        # reset member error references to the new data values
        for _err_dict in self._error_dicts.values():
            _err_dict['err'].reference = self._data
        self._clear_total_error_cache()

    @property
    def err(self):
        """absolute total data uncertainties (one-dimensional :py:obj:`numpy.ndarray`)"""
        _total_error = self.get_total_error()
        return _total_error.error

    @property
    def cov_mat(self):
        """absolute data covariance matrix (:py:obj:`numpy.matrix`)"""
        _total_error = self.get_total_error()
        return _total_error.cov_mat

    @property
    def cov_mat_inverse(self):
        """inverse of absolute data covariance matrix (:py:obj:`numpy.matrix`), or ``None`` if singular"""
        _total_error = self.get_total_error()
        return _total_error.cov_mat_inverse

    @property
    def cor_mat(self):
        """absolute data correlation matrix (:py:obj:`numpy.matrix`)"""
        _total_error = self.get_total_error()
        return _total_error.cor_mat

    @property
    def data_range(self):
        """
        :return: the minimum and maximum value of the data
        """
        return np.amin(self.data), np.amax(self.data)


    # -- public methods

    def add_error(self, err_val,
                  name=None, correlation=0, relative=False):
        """
        Add an uncertainty source to the data container.
        Returns an error id which uniquely identifies the created error source.

        :param err_val: pointwise uncertainty/uncertainties for all data points
        :type err_val: float or iterable of float
        :param name: unique name for this uncertainty source. If ``None``, the name
                     of the error source will be set to a random alphanumeric string.
        :type name: str or ``None``
        :param correlation: correlation coefficient between any two distinct data points
        :type correlation: float
        :param relative: if ``True``, **err_val** will be interpreted as a *relative* uncertainty
        :type relative: bool
        :return: error name
        :rtype: str
        """
        return super(IndexedContainer, self).add_error(
            err_val=err_val,
            name=name,
            correlation=correlation,
            relative=relative,
            reference=self._get_error_reference
        )

    def add_matrix_error(self, err_matrix, matrix_type,
                         name=None, err_val=None, relative=False):
        """
        Add a matrix uncertainty source to the data container.
        Returns an error id which uniquely identifies the created error source.

        :param err_matrix: covariance or correlation matrix
        :param matrix_type: one of ``'covariance'``/``'cov'`` or ``'correlation'``/``'cor'``
        :type matrix_type: str
        :param name: unique name for this uncertainty source. If ``None``, the name
                     of the error source will be set to a random alphanumeric string.
        :type name: str or ``None``
        :param err_val: the pointwise uncertainties (mandatory if only a correlation matrix is given)
        :type err_val: iterable of float
        :param relative: if ``True``, the covariance matrix and/or **err_val** will be interpreted
                         as a *relative* uncertainty
        :type relative: bool
        :return: error name
        :rtype: str
        """
        return super(IndexedContainer, self).add_matrix_error(
            err_matrix=err_matrix,
            matrix_type=matrix_type,
            name=name,
            err_val=err_val,
            relative=relative,
            reference=self._get_error_reference
        )



r"""The :py:mod:`kafe2.fit` module provides an essential toolkit for estimating model
parameters from data ("fitting").

It distinguishes between a number of different data types:

*  series of indexed measurements (dedicated submodule: :py:mod:`~kafe2.fit.indexed`),
*  *xy* data (dedicated submodule: :py:mod:`~kafe2.fit.xy`), and
*  histograms (dedicated submodule: :py:mod:`~kafe2.fit.histogram`)

Each of the above data types has its own particularities when it comes to fitting.
The main difference is due to the way uncertainties can be defined and
interpreted for each type of data.


Indexed data
------------

For **indexed** data, one data set consists of a list of :math:`N` distinct measurements
:math:`d_i`, with the (discrete) index :math:`i` ranging from :math:`0` to :math:`N-1`.
For each measurement in the series, one or more uncertainty sources can be defined,
each being a numerical estimate of how much the respective measurement fluctuates.
Correlations between uncertainties on separate measurements :math:`d_i` and :math:`d_j` can also
be taken into account by using *covariance/correlation matrices*.

Fits to *indexed* data take these uncertainties and correlations into account when estimating
the model parameters and their uncertainties. When plotting indexed data, measurements are
represented as data points with error bars at :math:`(x=i, y=d_i)`, whereas the model
is indicated by a horizontal line near the corresponding data point.

The following objects are provided for handling *indexed* data, as described above:

*  :py:obj:`~kafe2.fit.indexed.IndexedContainer`: data container for storing *indexed* data
*  :py:obj:`~kafe2.fit.indexed.IndexedParametricModel`: corresponding model:

   - uses a model function (:py:obj:`~kafe2.fit.indexed.model.IndexedModelFunction`) to calculate the model
     predictions and stores the result in an :py:obj:`~kafe2.fit.IndexedContainer`

*  :py:obj:`~kafe2.fit.indexed.IndexedFit`: a fit of a parametric model to *indexed* data:

   - finds the minimum of the *cost function* to find the
     parameter values for which the model best fits the data


XY data
-------

For **xy** data, the same principle as for *indexed* data applies, except each measurement and model prediction
now depends on a continuous real independent variable :math:`x` instead of a discrete index :math:`i`.
In effect, the data now consist of :math:`N` ordered pairs :math:`(x=x_i, y=d_i)`.

In contrast to *indexed* data, where only uncertainties on the measurement could be defined,
for *xy* data there is the additional possibility of defining additional uncertainites on :math:`x`.
These can be handled in a number of different ways when fitting an *xy* model to data.
When plotting the result of *xy* fits, the model function is displayed as a continuous
function of :math:`x`, and an **error band** can be computed to reflect the model uncertainty,
as determined by propagating the data uncertainties.

.. TODO: complete section on x errors

The following objects are provided for handling *xy* data:

*  :py:obj:`~kafe2.fit.xy.XYContainer`: data container for storing *xy* data
*  :py:obj:`~kafe2.fit.xy.XYParametricModel`: corresponding model:

   - uses a model function (:py:obj:`~kafe2.fit.xy.model.XYModelFunction`) to calculate the model predictions and
     stores the result in an :py:obj:`~kafe2.fit.XYContainer`

*  :py:obj:`~kafe2.fit.xy.XYFit`: a fit of a parametric model to *xy* data:

   - finds the minimum of the *cost function* to find the parameter values for which the model best fits the data


Histograms
----------

Finally, *kafe2* is also able to handle **histograms**. Histograms organize measurements whose
values can fall anywhere across a continuum of values into a number of discrete regions
or "bins". Typically, the continuous "measurement space" (a closed real interval :math:`[x_{\rm min}, x_{\rm max}]`)
is subdivided into a sequence of successive intervals at the "bin edges"
:math:`x_{\rm min} < x_1 < x_2 < \ldots < x_{\rm max}`. Whenever a measurement falls into one of the bins, the
value of that histogram bin is incremented by one.
So a histogram is completely defined by its bin edges and the bin values.

.. note::
    The bin numbering starts at :math:`1` for the first bin and ends at :math:`N`, where :math:`N`
    is defined as the *size* of the histogram. The bin numbers :math:`0` and :math:`N+1` refer to the
    underflow (below :math:`x_{\rm min}`) and overflow bin (above :math:`x_{\rm max}`), respectively.


Defining a parametric model for histograms is not as straightforward as for *xy* and *indexed* data.
Seeing as they keep track of the number of entries in different intervals of the continuum, the bin values
can be interpreted using probability theory.

As the number of entries approaches infinity, the number of entries :math:`n` in the bin covering an interval
:math:`[a, b)`, divided by the total number of entries :math:`N_{\rm E}`, will approach the probablity of an
event landing in that bin:

.. math::
    \lim_{N_{\rm E}\rightarrow\infty} \frac{n}{N_{\rm E}} = \int_a^b f(x)\,{\rm d}x = F(b) - F(a)

In the above formula, :math:`f(x)` is the **probability distribution function** (or **probability density**),
and :math:`F(x)` is an antiderivative of :math:`f`
(for example the **cumulative probability distribution function**).

Using the above relation, the model prediction :math:`m` for the bin :math:`[a, b)` can be defined as:

.. math::
    m = N_{\rm E} \int_a^b f(x)\,{\rm d}x = N_{\rm E} \left(F(b) - F(a)\right)

This means that, for histograms, the *model density* :math:`f(x)` needs to be specified.
The model is then calculated by numerically integrating this function over each bin.
When fitting, however, the model needs to be calculated for many different points in parameter space,
which makes this approach very slow (many numerical integrations until the fit converges).

An alternative would be to specify the model density *antiderviative* :math:`F` alongside
the model, so that the model can be calculated as a simple difference, rather than a complicated integral.

The following objects are provided for handling histograms:

*  :py:obj:`~kafe2.fit.histogram.HistContainer`: data container for storing histograms
*  :py:obj:`~kafe2.fit.histogram.HistParametricModel`: corresponding model:

   - uses a model function (:py:obj:`~kafe2.fit.xy.model.HistModelFunction`) to calculate the model predictions
     and stores the result in an :py:obj:`~kafe2.fit.HistContainer`

*  :py:obj:`~kafe2.fit.histogram.HistFit`: a fit of a parametric model to histograms:

   - finds the minimum of the *cost function* to find the parameter values for which the model best fits the data

For creating graphical representations of fits, the :py:obj:`~kafe2.fit.Plot` is provided. It can be instantiated
with any fit object (or list of fit objects) as an argument and will produce one or more plots accordingly
using `matplotlib`.

:synopsis: This module contains specialized objects for storing measurement data,
           defining and fitting parametric models to these data and producing
           graphical representations ("plots") of the result.
           It relies on the :py:mod:`kafe2.core` module for basic functionality.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""


# public interface of submodule 'kafe2.fit'


from .custom import *
from .indexed import *
from .histogram import *
from .multi import *
from .tools import *
from .xy import *
from .unbinned import *

from ._base.plot import Plot
from .tools.fit_wrapper import Fit  # import after every fit to avoid import conflicts with other tools

from .util import function_library, xy_fit, plot, plot_xy_data, k2Fit



import numpy as np
from scipy.optimize import minimize


def calculate_expected_fit_parameters_xy(
        x_data, y_data, model_function, y_error, initial_parameter_values, x_error=None,
        model_function_derivative=None, relative_model_y_error=False):
    x_data = np.array(x_data)
    y_data = np.array(y_data)
    y_error = np.array(y_error)
    initial_parameter_values = np.array(initial_parameter_values)

    def chi2(parameter_values):
        _residuals = y_data - model_function(x_data, *parameter_values)
        return np.sum((_residuals / y_error) ** 2) + np.sum(np.log(y_error ** 2))

    def chi2_dynamic_error(parameter_values):
        _model = model_function(x_data, *parameter_values)
        _residuals = y_data - _model
        _x_error = x_error * model_function_derivative(x_data, *parameter_values) \
            if x_error is not None else 0
        _x_error = 0
        _y_error = y_error if not relative_model_y_error else y_error * _model
        _xy_error_squared = _x_error ** 2 + _y_error ** 2
        return np.sum(_residuals ** 2 / _xy_error_squared) + np.sum(np.log(_xy_error_squared))

    _result = minimize(fun=chi2, x0=initial_parameter_values)
    print(1, _result.x)
    if x_error is None and not relative_model_y_error:
        return _result.x
    _result_2 = minimize(fun=chi2_dynamic_error, x0=_result.x)
    print(2, _result_2.x)
    return _result_2.x


def calculate_expected_multi_xy_chi2_cost(
        y_data_1, y_data_2, y_model_1, y_model_2, y_cov_mat_1, y_cov_mat_2, y_cov_mat_shared,
        x_cov_mat_1, x_cov_mat_2, x_cov_mat_shared, derivatives_1, derivatives_2):

    assert y_data_1.ndim == 1
    _shape_1d = y_data_1.shape
    _middle = _shape_1d[0]
    _end = 2 * _middle
    _shape_2d = 2 * _shape_1d
    assert y_data_2.shape == _shape_1d
    assert y_model_1.shape == _shape_1d
    assert y_model_2.shape == _shape_1d
    assert y_cov_mat_1.shape == _shape_2d
    assert y_cov_mat_2.shape == _shape_2d
    assert y_cov_mat_shared.shape == _shape_2d
    assert x_cov_mat_1.shape == _shape_2d
    assert x_cov_mat_2.shape == _shape_2d
    assert x_cov_mat_shared.shape == _shape_2d
    assert derivatives_1.shape == _shape_1d
    assert derivatives_2.shape == _shape_1d

    _residuals = np.zeros(shape=_end)
    _residuals[0:_middle] = y_data_1 - y_model_1
    _residuals[_middle:_end] = y_data_2 - y_model_2

    _xy_cov_mat = np.zeros(shape=(_end, _end))

    # Add y cov mats first
    _xy_cov_mat[0:_middle, 0:_middle] += y_cov_mat_1
    _xy_cov_mat[0:_middle, 0:_middle] += y_cov_mat_shared
    _xy_cov_mat[_middle:_end, _middle:_end] += y_cov_mat_2
    _xy_cov_mat[_middle:_end, _middle:_end] += y_cov_mat_shared
    _xy_cov_mat[0:_middle, _middle:_end] += y_cov_mat_shared
    _xy_cov_mat[_middle:_end, 0:_middle] += y_cov_mat_shared

    # Add x cov mats afterwards
    _outer_11 = np.outer(derivatives_1, derivatives_1)
    _outer_22 = np.outer(derivatives_2, derivatives_2)
    _outer_12 = np.outer(derivatives_1, derivatives_2)
    _outer_21 = np.outer(derivatives_2, derivatives_1)
    _xy_cov_mat[0:_middle, 0:_middle] += x_cov_mat_1 * _outer_11
    _xy_cov_mat[0:_middle, 0:_middle] += x_cov_mat_shared * _outer_11
    _xy_cov_mat[_middle:_end, _middle:_end] += x_cov_mat_2 * _outer_22
    _xy_cov_mat[_middle:_end, _middle:_end] += x_cov_mat_shared * _outer_22
    _xy_cov_mat[0:_middle, _middle:_end] += x_cov_mat_shared * _outer_12
    _xy_cov_mat[_middle:_end, 0:_middle] += x_cov_mat_shared * _outer_21

    _xy_cov_mat_inv = np.linalg.inv(_xy_cov_mat)

    return np.matmul(_residuals.T, np.matmul(_xy_cov_mat_inv, _residuals))



from ....fit import HistContainer, IndexedContainer, XYContainer, UnbinnedContainer
from .. import _AVAILABLE_REPRESENTATIONS
from ._base import DataContainerDReprBase
from .._yaml_base import YamlWriterMixin, YamlReaderMixin
from ..error import common_error_tools

__all__ = ["DataContainerYamlReader", "DataContainerYamlWriter"]


class DataContainerYamlWriter(YamlWriterMixin, DataContainerDReprBase):
    DUMPER = common_error_tools.MatrixYamlDumper

    def __init__(self, data_container, output_io_handle):
        super(DataContainerYamlWriter, self).__init__(output_io_handle=output_io_handle,
                                                      data_container=data_container)

    @classmethod
    def _make_representation(cls, container):
        """Create a dictionary representing a data container.

        :param container: The data container which will be converted.
        :type container: kafe2.fit._base.DataContainerBase
        """
        _yaml_doc = dict()
        _class = container.__class__

        # -- determine container type
        _type = cls._CLASS_TO_OBJECT_TYPE_NAME.get(_class, None)
        if _type is None:
            raise TypeError("Container type unknown or not supported: %s" % _class)
        _yaml_doc['type'] = _type

        # -- write representation for container types
        if _class is HistContainer:
            _yaml_doc['bin_edges'] = list(map(float, container.bin_edges))
            if container._manual_heights:
                _yaml_doc['bin_heights'] = list(map(float, container.data))  # float64 -> float
                _yaml_doc['underflow'] = float(container.underflow)
                _yaml_doc['overflow'] = float(container.underflow)
            else:
                _yaml_doc['raw_data'] = list(map(float, container.raw_data))  # float64 -> float
        elif _class is IndexedContainer or _class is UnbinnedContainer:
            _yaml_doc['data'] = container.data.tolist()
        elif _class is XYContainer:
            _yaml_doc['x_data'] = container.x.tolist()
            _yaml_doc['y_data'] = container.y.tolist()
        else:
            raise TypeError("Container type unknown or not supported: {}".format(_type))

        # -- write error representation for all container types
        if container.has_errors:
            common_error_tools.write_errors_to_yaml(container, _yaml_doc)

        # write labels for all container types
        if container.label is not None:
            _yaml_doc['label'] = container.label
        _x_axis_label, _y_axis_label = container.axis_labels
        if _x_axis_label is not None:
            _yaml_doc['x_label'] = _x_axis_label
        if _y_axis_label is not None:
            _yaml_doc['y_label'] = _y_axis_label

        return _yaml_doc


class DataContainerYamlReader(YamlReaderMixin, DataContainerDReprBase):
    LOADER = common_error_tools.MatrixYamlLoader

    def __init__(self, input_io_handle):
        super(DataContainerYamlReader, self).__init__(input_io_handle=input_io_handle,
                                                      data_container=None)

    @classmethod
    def _get_required_keywords(cls, yaml_doc, container_class):
        if container_class is HistContainer:
            return []
        if container_class is IndexedContainer or container_class is UnbinnedContainer:
            return ['data']
        if container_class is XYContainer:
            return ['x_data', 'y_data']
        raise TypeError("Unknown container type")

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        # -- determine container class from type
        _container_type = yaml_doc.pop('type')
        _class = cls._OBJECT_TYPE_NAME_TO_CLASS.get(_container_type)

        # -- read in representation for container types
        if _class is HistContainer:
            _bin_edges = yaml_doc.pop('bin_edges', None)
            _n_bins = yaml_doc.pop('n_bins', None)
            _bin_range = yaml_doc.pop('bin_range', None)
            if not _bin_edges and not (_n_bins and _bin_range):
                raise ValueError("When reading in a histogram dataset either "
                                 "bin_edges or n_bins and bin_range has to be specified!")
            if _bin_edges:
                _n_bins = len(_bin_edges) - 1
                _bin_range = (_bin_edges[0], _bin_edges[-1])
            _raw_data = yaml_doc.pop('raw_data', None)
            _bin_heights = yaml_doc.pop('bin_heights', None)
            if _raw_data and _bin_heights:
                raise ValueError("When reading in a histogram dataset only one out of "
                                 "raw_data and bin_heights can be specified!")
            _container_obj = HistContainer(n_bins=_n_bins,
                                           bin_range=_bin_range,
                                           bin_edges=_bin_edges,
                                           fill_data=_raw_data)
            if _bin_heights:
                _underflow = yaml_doc.pop('underflow', 0)
                _overflow = yaml_doc.pop('overflow', 0)
                _container_obj.set_bins(
                    bin_heights=_bin_heights, underflow=_underflow, overflow=_overflow)
        elif _class is IndexedContainer:
            _data = yaml_doc.pop('data')
            _container_obj = IndexedContainer(_data)
        elif _class is UnbinnedContainer:
            _data = yaml_doc.pop('data')
            _container_obj = UnbinnedContainer(_data)
        elif _class is XYContainer:
            _x_data = yaml_doc.pop('x_data')
            _y_data = yaml_doc.pop('y_data')
            _container_obj = XYContainer(_x_data, _y_data)
        else:
            raise TypeError("Container type unknown or not supported: {}".format(_container_type))

        # get labels for all container types
        _container_obj.label = yaml_doc.pop('label', None)
        _container_obj.axis_labels = (yaml_doc.pop('x_label', None), yaml_doc.pop('y_label', None))

        # -- process error sources
        _container_obj, yaml_doc = common_error_tools.process_error_sources(
            container_obj=_container_obj, yaml_doc=yaml_doc)

        return _container_obj, yaml_doc


# register the above classes in the module-level dictionary
DataContainerYamlReader._register_class(_AVAILABLE_REPRESENTATIONS)
DataContainerYamlWriter._register_class(_AVAILABLE_REPRESENTATIONS)



"""This submodule contains objects for handling different data representations. Whole fit objects can be saved and
loaded to and from a YAML file. Support for more file formats might follow in future updates.

:synopsis: This submodule contains objects for handling different data representations.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

# this module-level dict is populated by importing the
# individual modules below
_AVAILABLE_REPRESENTATIONS = dict()

# register different aliases (i.e. file extensions) for file formats
_FILE_FORMAT_ALIAS_RESOLVDICT = dict(yml='yaml')


from .constraint import *
from .container import *
from .fit import *
from .format import *
from .model import *


def _get_representer(object_type_name, role, file_format):
    _d_roles = _AVAILABLE_REPRESENTATIONS.get(object_type_name, None)
    if _d_roles is None:
        raise TypeError(
            "No representers found for object type '{}'! Available object types "
            "are: {}".format(object_type_name, _AVAILABLE_REPRESENTATIONS.keys()))

    _d_fileformats = _d_roles.get(role, None)
    if _d_fileformats is None:
        raise TypeError(
            "No representers with role '{}' found for object type '{}'! "
            "Available roles are: {}".format(role,
                                             object_type_name,
                                             _d_roles.keys()))

    _file_format_resolve_alias = _FILE_FORMAT_ALIAS_RESOLVDICT.get(file_format, file_format)
    _representer = _d_fileformats.get(_file_format_resolve_alias, None)
    if _representer is None:
        raise TypeError(
            "No representer with role '{}' found for of object type '{}' and file format '{}'! "
            "Available file formats are: {}".format(
                file_format,
                role,
                object_type_name, _d_fileformats.keys()))

    return _representer


# -- convenience functions

def get_reader(object_type_name, file_format):
    return _get_representer(object_type_name=object_type_name,
                            role='reader',
                            file_format=file_format)


def get_writer(object_type_name, file_format):
    return _get_representer(object_type_name=object_type_name,
                            role='writer',
                            file_format=file_format)



try:
    import typing  # help IDEs with type-hinting inside docstrings
except ImportError:
    pass
import numpy  # help IDEs with type-hinting inside docstrings
import numpy as np

from ...core.error import MatrixGaussianError, SimpleGaussianError
from ..indexed import IndexedContainer


__all__ = ['XYContainer']


class XYContainer(IndexedContainer):
    """
    This object is a specialized data container for *xy* data.

    """
    _AXIS_SPEC_DICT = {0: 0, 1: 1, '0': 0, '1': 1, 'x': 0, 'y': 1}
    
    # TODO Why does the XYContainer constructor require data while
    #      HistContainer and IndexedContainer don't?
    def __init__(self, x_data, y_data, dtype=float):
        """Construct a container for *xy* data:

        :param x_data: 1D array of measurement *x* values.
        :type x_data: typing.Sequence[dtype]
        :param y_data: 1D array of measurement *y* values.
        :type y_data: typing.Sequence[dtype]
        :param dtype: Data type of the measurements.
        :type dtype: type
        """
        # TODO: check user input (?)
        x_data = np.asarray(x_data)
        y_data = np.asarray(y_data)
        if x_data.ndim > 1:
            raise ValueError("x_data must be scalar or one-dimensional "
                             f"but received data with {x_data.ndim} dimensions.")
        if y_data.ndim > 1:
            raise ValueError("y_data must be scalar or one-dimensional "
                             f"but received data with {y_data.ndim} dimensions.")
        if x_data.shape != y_data.shape:
            raise ValueError("x_data and y_data must have the same shape but received "
                             f"{x_data.shape} for x_data and {y_data.shape} for y_data.")
        # super constructor doesn't allow 2D arrays
        super(XYContainer, self).__init__(np.zeros(len(x_data)))
        self._data = np.array([x_data, y_data], dtype=dtype)  # overwrite internal data storage

    # -- private methods

    @staticmethod
    def _find_axis_raise(axis_spec):
        try:
            axis_spec = axis_spec.lower()
        except AttributeError:
            # integers have no .lower() method
            pass
        _axis_id = XYContainer._AXIS_SPEC_DICT.get(axis_spec, None)
        if _axis_id is None:
            raise ValueError("No axis with id %r!" % (axis_spec,))
        return _axis_id

    def _get_data_for_axis(self, axis_id):
        return np.array(self._data[axis_id])

    def _get_error_reference(self, axis_id):
        return self._data[axis_id]

    def _calculate_total_error(self):
        _sz = self.size
        _tmp_cov_mat_x = np.zeros((_sz, _sz))
        _tmp_cov_mat_y = np.zeros((_sz, _sz))
        for _err_dict in self._error_dicts.values():
            if not _err_dict['enabled']:
                continue
            assert _err_dict['axis'] in (0, 1)
            if _err_dict['axis'] == 0:
                _tmp_cov_mat_x += _err_dict['err'].cov_mat
            elif _err_dict['axis'] == 1:
                _tmp_cov_mat_y += _err_dict['err'].cov_mat

        _total_err_x = MatrixGaussianError(_tmp_cov_mat_x, 'cov', relative=False, reference=self.x)
        _total_err_y = MatrixGaussianError(_tmp_cov_mat_y, 'cov', relative=False, reference=self.y)
        self._total_error = [_total_err_x, _total_err_y]

    def _clear_total_error_cache(self):
        """recalculate total errors next time they are needed"""
        self._total_error = None

    # -- public properties

    @property
    def size(self):
        """Number of data points.

        :rtype: int
        """
        return self._data.shape[1]

    @property
    def data(self):
        """2D array with shape ``(2, size)`` containing a copy of the data stored in this
        container.

        :rtype: numpy.ndarray
        """
        return self._data.copy()  # copy to ensure no modification by user

    @data.setter
    def data(self, new_data):
        _new_data = np.asarray(new_data)
        if _new_data.ndim != 2:
            raise ValueError("XYContainer data must be 2-d array of floats! "
                             "Got shape: %r..." % (_new_data.shape,))
        if _new_data.shape[0] == 2:
            self._data = _new_data.copy()
        elif _new_data.shape[1] == 2:
            self._data = _new_data.T.copy()
        else:
            raise ValueError("XYContainer data length must be 2 in at least one axis! "
                             "Got shape: %r..." % (_new_data.shape,))
        self._clear_total_error_cache()

    @property
    def x(self):
        """1D array of length ``[size]`` containing the *x* data.

        :rtype: numpy.ndarray
        """
        return self._get_data_for_axis(0)

    @x.setter
    def x(self, new_x):
        _new_x_data = np.squeeze(np.array(new_x))
        if len(_new_x_data.shape) > 1:
            raise ValueError("XYContainer 'x' data must be 1-d array of floats! "
                             "Got shape: %r..." % (_new_x_data.shape,))
        self._data[0, :] = new_x
        for _err_dict in self._error_dicts.values():
            if _err_dict['axis'] == 0:
                _err_dict['err'].reference = self._get_data_for_axis(0)
        self._clear_total_error_cache()

    @property
    def x_err(self):
        """1D array containing the absolute total data *x* uncertainties.

        :rtype: numpy.ndarray
        """
        _total_error_x = self.get_total_error(axis=0)
        return _total_error_x.error

    @property
    def x_cov_mat(self):
        """2D array of shape ``(size, size)`` containing the absolute data *x* covariance matrix.

        :rtype: numpy.ndarray"""
        _total_error_x = self.get_total_error(axis=0)
        return _total_error_x.cov_mat

    @property
    def x_cov_mat_inverse(self):
        """2D array of shape ``(size, size)`` containing the inverse of the absolute data *x*
        covariance matrix. :py:obj:`None` if singular.

        :rtype: numpy.ndarray or None
        """
        _total_error_x = self.get_total_error(axis=0)
        return _total_error_x.cov_mat_inverse

    @property
    def x_cor_mat(self):
        """2D array of shape ``(size, size)`` containing the absolute data *x* correlation matrix.

        :rtype: numpy.ndarray
        """
        _total_error_x = self.get_total_error(axis=0)
        return _total_error_x.cor_mat

    @property
    def y(self):
        """1D array of length ``size`` containing the *y* data.

        :rtype: numpy.ndarray
        """
        return self._get_data_for_axis(1)

    @y.setter
    def y(self, new_y):
        _new_y_data = np.squeeze(np.array(new_y))
        if len(_new_y_data.shape) > 1:
            raise ValueError("XYContainer 'y' data must be 1-d array of floats! "
                             "Got shape: %r..." % (_new_y_data.shape,))
        self._data[1, :] = new_y
        for _err_dict in self._error_dicts.values():
            if _err_dict['axis'] == 1:
                _err_dict['err'].reference = self._get_data_for_axis(1)
        self._clear_total_error_cache()

    @property
    def y_err(self):
        """1D array of length ``size`` containing the absolute total data *y* uncertainties.

        :rtype: numpy.ndarray
        """
        _total_error_y = self.get_total_error(axis=1)
        return _total_error_y.error

    @property
    def y_cov_mat(self):
        """2D array of shape ``(size, size)`` containing the absolute data *y* covariance matrix.

        :rtype: numpy.ndarray
        """
        _total_error_y = self.get_total_error(axis=1)
        return _total_error_y.cov_mat

    @property
    def y_cov_mat_inverse(self):
        """2D array of shape ``(size, size)`` containing the inverse of absolute data *y*
        covariance matrix. :py:obj:`None` if singular.

        :rtype: numpy.ndarray
        """
        _total_error_y = self.get_total_error(axis=1)
        return _total_error_y.cov_mat_inverse

    @property
    def y_cor_mat(self):
        """2D array of shape ``(size, size)`` containing the absolute data *y* correlation
        matrix.

        :rtype: numpy.ndarray
        """
        _total_error_y = self.get_total_error(axis=1)
        return _total_error_y.cor_mat

    @property
    def x_range(self):
        """Minimum and maximum values of the *x* data.

        :rtype: tuple[float, float]
        """
        _x = self.x
        return np.min(_x), np.max(_x)

    @property
    def y_range(self):
        """Minimum and maximum values of the *y* data.

        :rtype: tuple[float, float]
        """
        _y = self.y
        return np.min(_y), np.max(_y)

    # -- public methods

    def add_error(self, axis, err_val, name=None, correlation=0, relative=False):
        """Add an uncertainty source for an axis to the data container.

        :param axis: ``'x'``/``0`` or ``'y'``/``1``
        :type axis: str or int
        :param err_val: Pointwise uncertainties or a single uncertainty for all data points.
        :type err_val: float or typing.Sequence[float]
        :param name: Unique name for this uncertainty source. If :py:obj:`None`, the name
            of the error source will be set to a random alphanumeric string.
        :type name: str or None
        :param correlation: Correlation coefficient between any two distinct data points.
        :type correlation: float
        :param relative: If :py:obj:`True`, **err_val** will be interpreted as a *relative*
            uncertainty.
        :type relative: bool
        :return: An error id uniquely identifying the created error source.
        :rtype: str
        """
        _axis = self._find_axis_raise(axis)
        err_val = np.asarray(err_val, dtype=float)

        if err_val.ndim == 0:  # if dimensionless numpy array (i.e. float64), add a dimension
            err_val = np.ones(self.size) * err_val

        _err = SimpleGaussianError(
            err_val=err_val, corr_coeff=correlation, relative=relative,
            reference=lambda: self._get_error_reference(_axis)
        )
        _name = self._add_error_object(name=name, error_object=_err, axis=_axis)
        return _name

    def add_matrix_error(self, axis, err_matrix, matrix_type, name=None, err_val=None, relative=False):
        """Add a matrix uncertainty source for an axis to the data container.

        :param axis: ``'x'``/``0`` or ``'y'``/``1``
        :type axis: str or int
        :param err_matrix: 2D array of shape ``(size, size)`` containing the covariance or
            correlation matrix
        :type err_matrix: numpy.ndarray
        :param matrix_type: One of ``'covariance'``/``'cov'`` or ``'correlation'``/``'cor'``.
        :type matrix_type: str
        :param name: Unique name for this uncertainty source. If :py:obj:`None`, the name
            of the error source will be set to a random alphanumeric string.
        :type name: str or None
        :param err_val: The pointwise uncertainties. This is mandatory if only a correlation
            matrix is given.
        :type err_val: typing.Sequence[float]
        :param relative: If :py:obj:`True`, the covariance matrix and/or **err_val** will be
            interpreted as a *relative* uncertainty.
        :type relative: bool
        :return: An error id uniquely identifying the created error source.
        :rtype: str
        """
        _axis = self._find_axis_raise(axis)
        _err = MatrixGaussianError(
            err_matrix=err_matrix, matrix_type=matrix_type, err_val=err_val,
            relative=relative, reference=lambda: self._get_error_reference(axis)
        )
        _err.check_cov_mat_symmetry()
        _name = self._add_error_object(name=name, error_object=_err, axis=_axis)
        return _name

    def get_total_error(self, axis):
        """Get the error object representing the total uncertainty for a specific axis.

        :param axis: ``'x'``/``0`` or ``'y'``/``1``
        :type axis: str or int
        :return: Error object representing the total uncertainty.
        :rtype: kafe2.core.error.MatrixGaussianError
        """
        _axis = self._find_axis_raise(axis)
        if self._total_error is None:
            self._calculate_total_error()
        return self._total_error[_axis]

    @property
    def has_x_errors(self):
        """:py:obj:`True` if at least one *x* uncertainty source is defined for the data container.

        :rtype: bool
        """
        for _err_dict in self._error_dicts.values():
            if _err_dict['axis'] == 0:
                return True
        return False

    @property
    def has_uncor_x_errors(self):
        """:py:obj:`True` if at least one *x* uncertainty source, which is not fully correlated,
        is defined for the data container.

        :rtype: bool
        """
        for _err_dict in self._error_dicts.values():
            if _err_dict['axis'] == 0 and _err_dict['err'].corr_coeff != 1.0:
                return True
        return False

    @property
    def has_y_errors(self):
        """:py:obj:`True` if at least one *x* uncertainty source is defined for the data container.

        :rtype: bool
        """
        for _err_dict in self._error_dicts.values():
            if _err_dict['axis'] == 1:
                return True
        return False



import six
import sys
from .._base import FitBase, CostFunction, ModelFunctionBase
from ...core.fitters.nexus import Parameter, Array


class CustomFit(FitBase):
    _MODEL_ERROR_NODE_NAMES = []

    def __init__(
            self, cost_function, minimizer=None, minimizer_kwargs=None):
        if not isinstance(cost_function, CostFunction):
            cost_function = CostFunction(
                cost_function, add_constraint_cost=True, add_determinant_cost=False)
            cost_function._needs_errors = False
        super().__init__(
            data=None, model_function=None, cost_function=cost_function,
            minimizer=minimizer, minimizer_kwargs=minimizer_kwargs,
            dynamic_error_algorithm="nonlinear"
        )
        _dummy_model_function = ModelFunctionBase(cost_function.func, independent_argcount=0)
        _parameter_nodes = []
        for _par_name, _par_value in six.iteritems(_dummy_model_function.defaults_dict):
            _parameter_nodes.append(self._nexus.add(Parameter(_par_value, name=_par_name),
                                                    existing_behavior="replace"))
            self._fit_param_names.append(_par_name)
        self._nexus.add(Array(_parameter_nodes, name="parameter_values"),
                        existing_behavior="replace")
        self._parameter_formatters = _dummy_model_function.formatter.arg_formatters
        self._initialize_fitter()
        """
        Construct a fit without explicit data and model from only a cost function.
        
        :param cost_function: the function to minimize.
        :type cost_function: callable
        :param minimizer: Name of the minimizer to use.
        :type minimizer: str or None
        :param minimizer_kwargs: Dictionary wit keywords for initializing the minimizer.
        :type minimizer_kwargs: dict or None
        """

    def _set_new_data(self, new_data):
        pass

    def _set_new_parametric_model(self):
        pass

    def _set_data_as_model_ref(self):
        pass

    def _get_model_function_argument_formatters(self):
        return self._parameter_formatters

    def _get_model_function_parameter_formatters(self):
        return self._parameter_formatters

    def _iterative_fits_needed(self):
        return False

    def _second_fit_needed(self):
        return False

    @property
    def data(self):
        return None

    @property
    def data_error(self):
        return None

    @property
    def data_cov_mat(self):
        return None

    @property
    def data_cov_mat_inverse(self):
        return None

    @property
    def data_cor_mat(self):
        return None

    @property
    def model(self):
        return None

    @property
    def model_error(self):
        return None

    @property
    def model_cov_mat(self):
        return None

    @property
    def model_cov_mat_inverse(self):
        return None

    @property
    def model_cor_mat(self):
        return None

    @property
    def total_error(self):
        return None

    @property
    def total_cov_mat(self):
        return None

    @property
    def total_cov_mat_inverse(self):
        return None

    @property
    def total_cor_mat(self):
        return None

    @property
    def model_label(self):
        return None

    @property
    def data_size(self):
        return None

    @property
    def model_count(self):
        return 0

    @property
    def ndf(self):
        return None

    @property
    def goodness_of_fit(self):
        return None

    @property
    def chi2_probability(self):
        return None

    def report(self, output_stream=sys.stdout, asymmetric_parameter_errors=False):
        super().report(
            output_stream=output_stream, show_data=False, show_model=False, show_fit_results=True,
            asymmetric_parameter_errors=asymmetric_parameter_errors)


"""This submodule provides the necessary objects for parameter estimation using data consisting of an indexed series of
measurements. This can be useful for calculating weighted mean values or template fits.

:synopsis: This submodule provides the necessary objects for parameter estimation using data consisting of an indexed
    series of measurements.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

from .container import *
from .cost import *
from .fit import *
from .format import *
from .model import *
from .plot import *



import warnings
import numpy as np

from ..indexed import IndexedContainer


__all__ = ['HistContainer']


class HistContainer(IndexedContainer):
    """
    This object is a specialized data container for organizing data into *histograms*.

    A histogram is a compact representation of a potentially large number of
    entries which are distributed along a continuum of values. Histograms
    divide the continuum into intervals ("bins") and count the number of entries
    per interval.

    .. .. inheritance-diagram:: HistContainer
    ..    :parts: 1

    """
    def __init__(self, n_bins, bin_range, bin_edges=None, fill_data=None, dtype=int):
        """
        Construct a histogram:

        :param n_bins: number of bins
        :type n_bins: int
        :param bin_range: the lower and upper edges of the entire histogram
        :type bin_range: tuple of floats
        :param bin_edges: the bin edges (if ``None``, each bin will have the same width)
        :type bin_edges: list of floats
        :param fill_data: entries to fill into the histogram
        :type fill_data: list of floats
        :param dtype: data type of histogram entries
        :type dtype: type
        """
        super(HistContainer, self).__init__(data=np.zeros(n_bins+2), dtype=dtype)  # underflow and overflow bins
        self._manual_heights = False
        self._processed_entries = []
        self._unprocessed_entries = []
        # TODO: think of a way to implement weights

        if len(bin_range) != 2:
            raise ValueError(
                "Invalid bin range specification: %r! Must be tuple of 2 floats..." % (bin_range,))
        low, high = tuple(bin_range)

        if bin_edges is not None:
            # user specified bin edges -> check that these match the given 'n_bins' and 'bin_range'
            _sz = self.size
            if len(bin_edges) == _sz + 1:
                # assume bin edges given including bin range
                if not (bin_edges[0] == low and bin_edges[-1] == high):
                    raise ValueError(
                        "Invalid bin edge specification! First and last elements %r must match histogram bin range %r."
                        % ((bin_edges[0], bin_edges[-1]), bin_range))
                _bin_edges_including_low_and_high = bin_edges
            elif len(bin_edges) == _sz - 1:
                # <check if ordered and strictly within bin range>
                _bin_edges_including_low_and_high = [low] + bin_edges + [high]
            else:
                raise ValueError(
                    "Invalid bin edge specification! Number of elements (%d) must match histogram bin number (%d) +/- 1."
                    % (len(bin_edges), _sz))
            self.rebin(new_bin_edges=_bin_edges_including_low_and_high)
        else:
            # construct own bin edges
            self._bin_edges = np.linspace(bin_range[0], bin_range[1], n_bins+1)

        if fill_data is not None:
            self.fill(fill_data)

    # -- private methods

    def _fill_unprocessed(self):
        """fill any entries marked as unprocessed into the histogram"""
        if self._manual_heights:
            raise RuntimeError("The bin heights have been set manually. Filling entries is not "
                               "available anymore. Please construct a new HistContainer!")
        if not self._unprocessed_entries:
            return
        _entries_sorted = np.sort(self._unprocessed_entries)

        _current_entry_index = 0
        _current_entry_value = _entries_sorted[0]
        _current_bin_index = 0  # start with underflow bin
        _current_bin_upper_edge = self.low
        _current_bin_lower_edge = -np.inf

        # print "Bin %d (%5.3f, %5.3f)" % (_current_bin_index, _current_bin_lower_edge, _current_bin_upper_edge)
        while _current_bin_upper_edge <= self.high:
            if _current_entry_value >= _current_bin_upper_edge:
                # if last bin, quit
                if _current_bin_upper_edge == self.high:
                    # print "\tthis is the last bin: quitting loop"
                    break
                # move to next bin
                _current_bin_index += 1
                _current_bin_upper_edge = self._bin_edges[_current_bin_index]
                _current_bin_lower_edge = self._bin_edges[_current_bin_index - 1]
                # print "\tmove to next bin"
                # print "Bin %d (%5.3f, %5.3f)" % (_current_bin_index, _current_bin_lower_edge, _current_bin_upper_edge)
                continue
            # increment histogram bin and move to next entry
            # print "\tputting entry #%d '%s' in bin %d (%5.3f, %5.3f)" % (
            #     _current_entry_index, _current_entry_value, _current_bin_index, _current_bin_lower_edge, _current_bin_upper_edge)
            # add to processed entries
            self._processed_entries.append(_current_entry_value)
            self._data[_current_bin_index] += 1
            _current_entry_index += 1
            if _current_entry_index >= len(_entries_sorted):    # important BUGFIX!!!!!!!!!!
                break
            _current_entry_value = _entries_sorted[_current_entry_index]

        # any remaining entries are overflows
        if _current_entry_index < len(_entries_sorted):    # important BUGFIX!!!!!!!!!!

            _overflow_entries = _entries_sorted[_current_entry_index:]
            self._data[-1] += len(_overflow_entries)
            # print "Bin %d (of): add %d entries" % (_current_bin_index+1, len(_overflow_entries))
            # for i in _overflow_entries:
            #     print '\t', i
            self._processed_entries += list(_overflow_entries)

        self._unprocessed_entries = []
        if np.all(self._data == np.floor(self._data)):
            warnings.warn("Historam data is int. Make sure to consider this for your bin edges!")

    def _get_error_reference(self):
        return self._data[1:-1]

    # -- public properties

    @property
    def size(self):
        """the number of bins (excluding underflow and overflow bins)"""
        return len(self._data) - 2  # don't consider underflow and overflow bins

    @property
    def n_entries(self):
        """the number of entries"""
        return np.sum(self._data) + len(self._unprocessed_entries)

    @property
    def data(self):
        """the number of entries in each bin"""
        if self._unprocessed_entries:  # process outstanding entries
            self._fill_unprocessed()
        # NOTE: returned array starts at 0
        return self._data[1:-1].copy()  # don't consider underflow and overflow bins

    @data.setter
    def data(self, data):
        raise TypeError("Changing histogram data directly is not allowed! Use fill() or set_bins().")

    @property
    def raw_data(self):
        """the number of entries in each bin"""
        # TODO: commit unprocessed entries (?)
        return self._processed_entries + self._unprocessed_entries

    @property
    def low(self):
        """the lower edge of the histogram"""
        return self._bin_edges[0]

    @property
    def high(self):
        """the upper edge of the histogram"""
        return self._bin_edges[-1]

    @property
    def bin_range(self):
        """a tuple containing the lower and upper edges of the histogram"""
        return self.low, self.high

    @property
    def overflow(self):
        """the number of entries in the overflow bin"""
        return self._data[-1]

    @property
    def underflow(self):
        """the number of entries in the underflow bin"""
        return self._data[0]

    @property
    def n_bins(self):
        """the number of bins"""
        return len(self._bin_edges) - 1

    @property
    def bin_edges(self):
        """a list of the bin edges (including the outermost ones)"""
        return self._bin_edges.copy()

    @property
    def bin_widths(self):
        """a list of the bin widths"""
        _be = self._bin_edges
        return (_be[1:] - _be[:-1])

    @property
    def bin_centers(self):
        """a list of the (geometrical) bin centers"""
        _be = self._bin_edges
        return 0.5*(_be[1:] + _be[:-1])

    # -- public methods

    def fill(self, entries):
        """
        Fill new entries into the histogram.

        :param entries: list of entries
        :type entries: list of floats
        """
        if self._manual_heights:
            raise RuntimeError("The bin heights have been set manually. Filling additional data is not "
                               "possible anymore. Please construct a new HistContainer!")
        if np.asarray(entries).ndim > 1:
            raise ValueError("Fill data must be scalar or one-dimensional "
                             f"but received fill data with {np.asarray(entries).ndim} dimensions.")
        try:
            self._unprocessed_entries += list(entries)
        except TypeError:
            self._unprocessed_entries.append(entries)

    def rebin(self, new_bin_edges):
        """
        Change the histogram binning.

        :param new_bin_edges: list of new bin edges in ascending order
        :type new_bin_edges: list of float
        """
        if self._manual_heights:
            raise RuntimeError("The bin heights have been set manually. Rebinning is not possible "
                               "anymore. Please construct a new HistContainer!")
        _new_bin_edges = np.asarray(new_bin_edges, dtype=float)
        # check if list is sorted
        if not (np.diff(_new_bin_edges) >= 0).all():
            raise ValueError(
                "Invalid bin edge specification! Edge sequence must be sorted in ascending order!")
        self._bin_edges = _new_bin_edges
        self._data = np.zeros(len(self._bin_edges) - 1 + 2)

        # mark all entries as unprocessed
        self._unprocessed_entries += self._processed_entries
        self._processed_entries = []

    def set_bins(self, bin_heights, underflow=0, overflow=0):
        """
        Set the bin heights according to a pre-calculated histogram
        :param bin_heights: Heights of the bins
        :type bin_heights: list of int
        :param underflow: Number of entries in the underflow bin
        :type underflow: int
        :param overflow: Number of entries in the overflow bin
        :type overflow: int
        """
        self._manual_heights = True
        _new_data = np.array(bin_heights)
        if _new_data.ndim != 1:
            raise ValueError('Invalid dimensions for bin heights. '
                             f'Got {_new_data.ndim}-d array, expected 1-d array')
        _new_data = np.append(np.insert(_new_data, 0, underflow), overflow)
        if len(_new_data) != len(self._data):
            raise ValueError('Length of bin entries does not match binning. '
                             'Got {}, expected {}'.format(len(_new_data)-2, len(self._data)-2))
        self._data = _new_data
        self._processed_entries = []
        self._unprocessed_entries = []



import abc
import six

from .._base import GenericDReprBase
# import ModelFunction classes
from ..._base import ModelFunctionBase
from ...histogram import HistModelFunction, HistParametricModel
from ...indexed import IndexedModelFunction, IndexedParametricModel
from ...unbinned import UnbinnedParametricModel
from ...xy import XYParametricModel

__all__ = ["ModelFunctionDReprBase", "ParametricModelDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class ModelFunctionDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = 'model_function'

    _CLASS_TO_OBJECT_TYPE_NAME = {
        HistModelFunction: 'histogram',
        IndexedModelFunction: 'indexed',
        ModelFunctionBase: 'base'
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        'histogram': HistModelFunction,
        'indexed': IndexedModelFunction,
        'unbinned': ModelFunctionBase,
        'xy': ModelFunctionBase,  # type from fit is passed to model function, needs to be resolved
        'base': ModelFunctionBase
    }

    def __init__(self, model_function=None):
        self._kafe_object = model_function
        super(ModelFunctionDReprBase, self).__init__()


@six.add_metaclass(abc.ABCMeta)
class ParametricModelDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = 'model'

    _CLASS_TO_OBJECT_TYPE_NAME = {
        HistParametricModel: 'histogram',
        IndexedParametricModel: 'indexed',
        UnbinnedParametricModel: 'unbinned',
        XYParametricModel: 'xy'
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        'histogram': HistParametricModel,
        'indexed': IndexedParametricModel,
        'unbinned': UnbinnedParametricModel,
        'xy': XYParametricModel
    }

    def __init__(self, parametric_model=None):
        self._kafe_object = parametric_model
        super(ParametricModelDReprBase, self).__init__()



"""This submodule contains objects for handling different representations of data containers.

:synopsis: This submodule contains objects for handling different representations of data containers.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

from ._base import *
from .yaml_drepr import *



#!/usr/bin/env python
import argparse
import sys

import matplotlib.pyplot as plt

# do not use relative imports here as the working directory of kafe2go can be anywhere on the system
from kafe2 import Plot
from kafe2.fit._base.fit import FitBase
from kafe2.fit.tools.contours_profiler import ContoursProfiler
from kafe2.fit.xy.plot import XYPlotAdapter


def kafe2go():
    _parser = argparse.ArgumentParser(
        description="Perform a fit with the kafe2 package driven by an input file.\n"
                    "Example files can be found at "
                    "https://github.com/dsavoiu/kafe2/tree/master/examples.\n"
                    "Further information on how to create input files is given at "
                    "https://kafe2.readthedocs.io/en/latest/parts/user_guide.html#kafe2go."
        )

    _parser.add_argument('filename', type=str, nargs='+',
                         help="Name(s) of fit input file(s).")
    _parser.add_argument('-if', '--inputformat',
                         type=str, default='yaml',
                         help="File input format. The default format is yaml.")
    _parser.add_argument('-s', '--saveplot',
                         action='store_true',
                         help="Save plot(s) to file(s). The plot(s) will be saved in the current "
                              "working directory.")
    _parser.add_argument('-pf', '--plotformat',
                         type=str, default='pdf',
                         help="Graphics output file format. E.g. pdf, png, svg, ... "
                              "The default format is pdf.")
    _parser.add_argument('-n', '--noplot',
                         action='store_true',
                         help="Don't show plots on screen.")
    _parser.add_argument('-r', '--ratio',
                         action='store_true',
                         help="Show data/model ratio below the main plot.")
    _parser.add_argument('-R', '--residual',
                         action='store_true',
                         help="Show residuals below the main plot.")
    _parser.add_argument('-a', '--asymmetric',
                         action='store_true',
                         help="Show asymmetric parameter uncertainties when displaying the fit "
                              "information. This affects the fit report to the terminal as well as "
                              "the information box of the plot.")
    _parser.add_argument('-c', '--contours',
                         action='store_true',
                         help="Plot contours and profiles.")
    _parser.add_argument('--grid', type=str, nargs=1, default=[None],
                         help="Add a grid to the contour profiles. Available options are either "
                              "all, contours or profiles.")
    _parser.add_argument('--noband',
                         action='store_true',
                         help="Don't draw the 1-sigma band around the fitted function. "
                              "This will only affect plots of XY-fits.")
    _parser.add_argument('--noinfobox',
                         action='store_true',
                         help="Don't add the model info boxes to the plot(s).")
    _parser.add_argument('--separate',
                         action='store_true',
                         help="Create a separate figure for each fit when plotting.")
    _parser.add_argument('--noreport',
                         action='store_true',
                         help="Don't print fit report(s) to the terminal after fitting.")

    if len(sys.argv) == 1:  # print help message if no input given
        _parser.print_help()
        sys.exit(1)
    _args = _parser.parse_args()

    _filenames = _args.filename
    _band = not _args.noband
    _contours = _args.contours
    _report = not _args.noreport
    _infobox = not _args.noinfobox
    _ratio = _args.ratio
    _residual = _args.residual
    _asymmetric = _args.asymmetric
    _separate = _args.separate
    _input_format = _args.inputformat
    _plot_format = _args.plotformat
    _save_plot = _args.saveplot
    _show_plot = not _args.noplot
    _grid = _args.grid[0]

    _fits = []

    for _fname in _filenames:
        _fit = FitBase.from_file(_fname, file_format=_input_format)
        _fit.do_fit()
        if _report:
            _fit.report(asymmetric_parameter_errors=_asymmetric)
        _fits.append(_fit)

    if not _band:
        XYPlotAdapter.PLOT_SUBPLOT_TYPES.pop('model_error_band')

    _plot = Plot(fit_objects=_fits, separate_figures=_separate)
    _plot.plot(
        fit_info=_infobox, asymmetric_parameter_errors=_asymmetric, ratio=_ratio, residual=_residual
    )

    _basenames = [name.rsplit('.', 1)[0] for name in _filenames]

    if _save_plot:
        if len(_plot.figures) == 1:
            names = ['_'.join(_basenames)]
        else:
            names = _basenames
        for fig, name in zip(_plot.figures, names):
            fig.savefig(fname='{}.{}'.format(name, _plot_format), format=_plot_format)

    if _contours:
        for _fit, name in zip(_fits, _basenames):
            _profiler = ContoursProfiler(_fit)
            _profiler.plot_profiles_contours_matrix(show_grid_for=_grid)
            if _save_plot:
                for i, fig in enumerate(_profiler.figures):
                    fig.savefig(fname='{}_contours_{}.{}'.format(name, i, _plot_format),
                                format=_plot_format)

    if _show_plot:
        plt.show()


if __name__ == '__main__':
    kafe2go()
